# Copyright (c) 2025 Marcin Zdun
# This code is licensed under MIT license (see LICENSE for details)

"""
The **proj_flow.ext.python.steps** defines steps for building, installing and
documenting.
"""

import importlib
import os

from proj_flow.api import env, release, step

from . import rtdocs


@step.register
class Install:
    name = "Install"

    def platform_dependencies(self):
        return ["python -m pip"]

    def run(self, config: env.Config, rt: env.Runtime) -> int:
        return rt.cmd("python", "-m", "pip", "install", rt.root.as_posix())


@step.register
class Build:
    name = "Build"

    def platform_dependencies(self):
        return ["python -m build"]

    def run(self, config: env.Config, rt: env.Runtime) -> int:
        build_main = importlib.import_module("build.__main__")
        build_main.main([], "proj-flow build")
        return 0


@step.register
class CheckTwine:
    name = "Check Twine"

    runs_after = ["Build"]

    def platform_dependencies(self):
        return ["twine"]

    def run(self, config: env.Config, rt: env.Runtime) -> int:
        filenames: list[str] = []
        root = rt.root / "dist"
        for root, dirnames, filenames in root.walk():
            dirnames[:] = []

        _, project = release.project_suites.find(lambda suite: suite.get_project(rt))
        archive_name = project and project.archive_name
        if archive_name:
            dot_suffix = f"{archive_name}."
            dash_suffix = f"{archive_name}-"
            filenames = [
                filename
                for filename in filenames
                if filename.startswith(dot_suffix) or filename.startswith(dash_suffix)
            ]

        if len(filenames) == 0:
            return 0

        return rt.cmd(
            "twine", "check", *((root / filename).as_posix() for filename in filenames)
        )



# Copyright (c) 2026 Marcin Zdun
# This code is licensed under MIT license (see LICENSE for details)


import copy
import json
from dataclasses import dataclass
from pathlib import Path
from typing import Any, Callable, cast

import chevron

from proj_flow.ext.webidl.model.builders import ExtAttrsContextBuilders, TypeReplacement

package_root = Path(__file__).parent.parent
templates_dir = package_root / "data" / "templates"


def load_package_template(name: str):
    path = templates_dir / f"{name}.mustache"
    return path.read_text(encoding="UTF-8")


@dataclass
class CMakeDirs:
    base_dir: Path
    project_source_dir: Path
    project_binary_dir: Path
    current_source_dir: Path
    current_binary_dir: Path

    @staticmethod
    def from_config_path(config: Path, project_binary_dir: Path):
        config = config.absolute()
        project_binary_dir = project_binary_dir.absolute()
        base_dir = config.parent
        project_source_dir = Path().absolute()
        current_source_dir = project_source_dir

        for parent in config.parents:
            cmake_lists = parent / "CMakeLists.txt"
            if cmake_lists.exists():
                current_source_dir = parent
                break

        current_binary_dir = project_binary_dir / current_source_dir.relative_to(
            project_source_dir
        )

        return CMakeDirs(
            base_dir=base_dir,
            project_source_dir=project_source_dir,
            project_binary_dir=project_binary_dir,
            current_source_dir=current_source_dir,
            current_binary_dir=current_binary_dir,
        )

    def filename_from_config(self, config: Path, ext: str):
        in_source = config.absolute().relative_to(self.current_source_dir)
        in_build = self.current_binary_dir / in_source
        return Path(in_build.as_posix() + ext)

    def dependency_from_config(self, config: Path):
        return self.filename_from_config(config, ".deps")

    def cmake_from_config(self, config: Path):
        return self.filename_from_config(config, ".cmake")

    def my_defines(self):
        return {
            "PROJECT_SOURCE_DIR": self.project_source_dir.as_posix(),
            "PROJECT_BINARY_DIR": self.project_binary_dir.as_posix(),
            "CMAKE_CURRENT_SOURCE_DIR": self.current_source_dir.as_posix(),
            "CMAKE_CURRENT_BINARY_DIR": self.current_binary_dir.as_posix(),
        }

    def prefix(self, filename: str):
        if "${" in filename:
            return Path(filename)

        full = self.base_dir / filename

        if full.is_relative_to(self.current_binary_dir):
            return Path("${CMAKE_CURRENT_BINARY_DIR}") / full.relative_to(
                self.current_binary_dir
            )
        if full.is_relative_to(self.project_binary_dir):
            return Path("${PROJECT_BINARY_DIR}") / full.relative_to(
                self.project_binary_dir
            )
        if full.is_relative_to(self.current_source_dir):
            return Path("${CMAKE_CURRENT_SOURCE_DIR}") / full.relative_to(
                self.current_source_dir
            )
        if full.is_relative_to(self.project_source_dir):
            return Path("${PROJECT_SOURCE_DIR}") / full.relative_to(
                self.project_source_dir
            )

        return Path(filename)

    def make_abs_path(self):
        def impl(filename: str):
            return self.prefix(filename)

        return impl


def _relative(path: Path, basedir: Path | None):
    if not basedir:
        return path
    return path.relative_to(basedir, walk_up=True)


class Rebuild:
    @staticmethod
    def __rebuild_map(data: dict, global_ctx: dict[str, str]):
        for key, value in data.items():
            data[key] = Rebuild.__rebuild_any(value, global_ctx)
        return data

    @staticmethod
    def __rebuild_list(data: list, global_ctx: dict[str, str]):
        for index, value in enumerate(data):
            data[index] = Rebuild.__rebuild_any(value, global_ctx)
        return data

    @staticmethod
    def __rebuild_str(data: str, global_ctx: dict[str, str]):
        return chevron.render(data, global_ctx)

    @staticmethod
    def __rebuild_any(data, global_ctx: dict[str, str]):
        if isinstance(data, str):
            return Rebuild.__rebuild_str(data, global_ctx)
        if isinstance(data, list):
            return Rebuild.__rebuild_list(data, global_ctx)
        if isinstance(data, dict):
            return Rebuild.__rebuild_map(data, global_ctx)

        return data

    @staticmethod
    def rebuild(initial_context: dict, global_ctx: dict[str, str]):
        return Rebuild.__rebuild_map(initial_context, global_ctx)


@dataclass
class ConfigOutput:
    output_name: str
    mustache_template: str | None
    output_template: str | None
    lang: str | None
    types: str | None
    partials: str | None
    initial_context: dict
    debug: bool
    ndebug: bool

    @staticmethod
    def from_config(
        data: dict[str, str | dict[str, Any]], root_default: "ConfigOutput"
    ):
        result = [ConfigOutput.from_dict(key, item) for key, item in data.items()]

        for index in range(len(result)):
            default = result[index]
            if default.output_name != "":
                continue

            del result[index]
            for item in result:
                item.update_from(default)

            break

        for item in result:
            item.update_from(root_default)

        return result

    @staticmethod
    def from_dict(
        output_name: str, data: str | dict[str, str | dict]
    ) -> "ConfigOutput":
        if isinstance(data, str):
            return ConfigOutput(
                output_name=output_name,
                mustache_template=data,
                output_template=None,
                lang=None,
                types=None,
                partials=None,
                initial_context={},
                debug=False,
                ndebug=False,
            )

        mustache_template = cast(str | None, data.get("template", None))
        output_template = cast(str | None, data.get("path", None))
        lang = cast(str | None, data.get("lang", None))
        types = cast(str | None, data.get("types", None))
        partials = cast(str | None, data.get("partials", None))
        initial_context = copy.deepcopy(cast(dict, data.get("context", {})))
        debug = not not data.get("debug", {})
        ndebug = not not data.get("ndebug", {})

        return ConfigOutput(
            output_name=output_name,
            mustache_template=mustache_template,
            output_template=output_template,
            lang=lang,
            types=types,
            partials=partials,
            initial_context=initial_context,
            debug=debug,
            ndebug=ndebug,
        )

    def update_from(self, default: "ConfigOutput"):
        self.mustache_template = self.mustache_template or default.mustache_template
        self.output_template = self.output_template or default.output_template
        self.lang = self.lang or default.lang
        self.types = self.types or default.types
        self.partials = self.partials or default.partials
        self.initial_context = {
            **copy.deepcopy(self.initial_context),
            **copy.deepcopy(default.initial_context),
        }
        self.debug = self.debug or default.debug
        self.ndebug = self.ndebug or default.ndebug


def _file_props(path: Path):
    suffixes = "".join(path.suffixes)
    name = path.name
    shortest = name[: -len(suffixes)]
    return {
        "filename": path.as_posix(),
        "dirname": path.parent.as_posix(),
        "basename": path.name,
        "stem": path.stem,
        "suffix": path.suffix,
        "suffixes": suffixes,
        "shortest_stem": shortest,
    }


def _rel_file_props(path: Path, root: Path):
    props: dict = _file_props(path)
    if path.is_relative_to(root):
        rel = path.relative_to(root)
        props["relative"] = {
            "filename": rel.as_posix(),
            "dirname": rel.parent.as_posix(),
        }
    return props


@dataclass
class Output:
    output: Path
    mustache_template: Path | None
    lang: str | None
    types: Path | None
    partials: Path | None
    initial_context: dict
    debug: bool

    @staticmethod
    def filename_context(filename: str, root_dir: Path):
        return _rel_file_props(Path(filename), root_dir)

    @staticmethod
    def abs_path(basedir: Path, path: str):
        return basedir / path

    @staticmethod
    def from_config(
        context: dict[str, str],
        root_default: ConfigOutput,
        data: dict[str, str | dict[str, Any]],
        abs_path: Callable[[str], Path],
    ):
        expand_path = lambda text: chevron.render(template=text, data=context)

        result: list[Output] = []

        for output in ConfigOutput.from_config(data, root_default):
            output_name = expand_path(output.output_name)
            context["PATH"] = str(output_name)

            if output.output_template:
                output_name = expand_path(output.output_template)
            output_name = abs_path(output_name)

            mustache_template = None
            if output.mustache_template:
                mustache_template = abs_path(expand_path(output.mustache_template))

            types = None
            if output.types:
                types = abs_path(expand_path(output.types))

            partials = None
            if output.partials:
                partials = abs_path(expand_path(output.partials))

            result.append(
                Output(
                    output=output_name,
                    mustache_template=mustache_template,
                    lang=output.lang,
                    types=types,
                    partials=partials,
                    initial_context=Rebuild.rebuild(output.initial_context, context),
                    debug=output.debug and not output.ndebug,
                )
            )

        return result

    def get_type_replacements(self):
        return TypeReplacement.load_config(self.lang, self.types)


@dataclass
class TemplateRule:
    inputs: list[Path]
    outputs: list[Output]

    @staticmethod
    def from_config(
        global_ctx: dict, source_dir: Path, data: dict, abs_path: Callable[[str], Path]
    ):
        templates = cast(dict[str, dict[str, str | dict]], data.get("templates", {}))
        inputs = cast(list[dict[str, str]], data.get("inputs", []))
        return cast(
            list[TemplateRule],
            list(
                filter(
                    lambda v: v is not None,
                    [
                        TemplateRule.__get_rule(
                            cast(list[str] | str, input.get("idl", [])),
                            cast(str | None, input.get("template")),
                            global_ctx,
                            source_dir,
                            templates,
                            abs_path,
                        )
                        for input in inputs
                    ],
                )
            ),
        )

    @staticmethod
    def __get_rule(
        input_names: str | list[str],
        suite_id: str | None,
        global_ctx: dict,
        source_dir: Path,
        templates: dict[str, dict[str, str | dict[str, Any]]],
        abs_path: Callable[[str], Path],
    ):
        if suite_id is None:
            return None

        names = [input_names] if isinstance(input_names, str) else input_names
        context = {**global_ctx, "input": Output.filename_context(names[0], source_dir)}

        inputs = [abs_path(name) for name in names]
        root_config = ConfigOutput.from_dict("", templates.get("", {}).get("", {}))
        outputs = Output.from_config(
            context, root_config, templates.get(suite_id, {}), abs_path
        )

        return TemplateRule(inputs=inputs, outputs=outputs)

    def get_dependencies(
        self, CMAKE_CURRENT_BINARY_DIR: Path | None, *additional_inputs: Path
    ) -> dict[str, list[str]]:
        def rel(path: Path):
            return _relative(path, CMAKE_CURRENT_BINARY_DIR)

        result: dict[str, list[str]] = {}
        for out in self.outputs:
            current_inputs: list[Path] = [*self.inputs, *additional_inputs]
            if out.mustache_template:
                current_inputs.append(out.mustache_template)
            result[rel(out.output).as_posix()] = [
                path.as_posix() for path in current_inputs
            ]
        return result


@dataclass
class TemplateConfig:
    version: int
    rules: list[TemplateRule]
    ext_attrs: ExtAttrsContextBuilders

    @staticmethod
    def load_config(
        config: Path,
        context: dict,
        source_dir: Path,
        abs_path: Callable[[str], Path] | None = None,
    ):
        with config.open(encoding="UTF-8") as json_file:
            data = cast(dict, json.load(json_file))

        config_dep = config.absolute()
        basedir = config_dep.parent

        return TemplateConfig(
            version=cast(int, data.get("version", 1)),
            rules=TemplateRule.from_config(
                context, source_dir, data, abs_path or (lambda path: basedir / path)
            ),
            ext_attrs=ExtAttrsContextBuilders.builtin().merge(
                ExtAttrsContextBuilders.from_config(data)
            ),
        )



# Copyright (c) 2024 Marcin Zdun
# This code is licensed under MIT license (see LICENSE for details)

"""
The **proj_flow.log.hosting.github** builds a log message for GiHub Releases.
"""

import json
from pprint import pformat
from typing import Any, List, NamedTuple, Optional, Tuple, cast

from proj_flow.api import env
from proj_flow.base import cmd
from proj_flow.log import commit, msg


class _GitHub(NamedTuple):
    remote: str
    owner: str
    repo: str


_NO_GITHUB = _GitHub("", "", "")


class ReleaseInfo(commit.ReleaseInfo):
    id: int

    def __init__(
        self,
        url: Optional[str],
        is_draft: Optional[bool],
        id: int,
        ref: Optional[str] = None,
        tag: Optional[str] = None,
    ):
        super().__init__(url, is_draft, ref, tag)
        self.id = id


class GitHub(commit.Hosting):
    """
    Generates links to GitHub.
    """

    _info: _GitHub
    rt: env.Runtime

    def __init__(self, info: _GitHub, rt: env.Runtime):
        super().__init__(f"https://github.com/{info.owner}/{info.repo}")
        self._info = info
        self.rt = rt

    @property
    def is_active(self):
        return not not self._info.remote

    @property
    def remote(self):
        return self._info.remote

    @property
    def owner(self):
        return self._info.owner

    @property
    def repo(self):
        return self._info.repo

    @property
    def root(self):
        return f"/repos/{self.owner}/{self.repo}"

    def single_commit_link(self, link: commit.Link) -> Optional[str]:
        return f"{self.host_link}/commit/{link.hash}"

    def commit_listing_link(self, setup: commit.LogSetup) -> Optional[str]:
        curr_tag = setup.curr_tag or "HEAD"
        if not setup.prev_tag:
            return f"{self.host_link}/commits/{curr_tag}"

        return f"{self.host_link}/compare/{setup.prev_tag}...{curr_tag}"

    def reference_link(self, ref: str) -> Optional[str]:
        if ref[:1] == "#" and ref[1:].isdigit():
            f"{self.host_link}/issues/{ref[1:]}"

        return None

    def gh(
        self,
        res: str,
        *args: str,
        method: Optional[str] = None,
        capture_output: bool = True,
        ro_call: bool = False,
        **kwargs,
    ):
        url = f"{self.root}{res}"
        accept = kwargs.get("accept", "application/vnd.github+json")
        call = [
            "gh",
            "api",
            "-H",
            f"Accept: {accept}",
            "-H",
            "X-GitHub-Api-Version: 2022-11-28",
        ]
        if method:
            call.extend(["--method", method.upper()])

        if capture_output:
            if not ro_call and self.rt.dry_run:
                self.rt.print(*call, url, *args)
                return None
            return self.rt.capture(*call, url, *args)

        return self._run(*call, url, *args)

    def _run(self, *args: str):
        self.rt.print(*args)
        if self.rt.dry_run:
            return None
        return cmd.run(*args)

    def json_from(
        self,
        res: str,
        *args: str,
        method: Optional[str] = None,
        server: Optional[str] = None,
        default: Any = {},
        ro_call: bool = False,
    ):
        proc = self.gh(res, *args, method=method, server=server, ro_call=ro_call)
        if proc is None:
            return default

        if self.rt.verbose:
            self.rt.message(f"[DEBUG] stderr={proc.returncode}")
            self.rt.message("[DEBUG] stderr=")
            try:
                self.rt.message(pformat(json.loads(proc.stderr)))
            except json.decoder.JSONDecodeError:
                self.rt.message(proc.stderr)
            self.rt.message("[DEBUG] stdout=")
            try:
                self.rt.message(pformat(json.loads(proc.stdout)))
            except json.decoder.JSONDecodeError:
                self.rt.message(proc.stdout)

        return json.loads(proc.stdout)

    def _release_from_json(self, data: dict, draft: bool = False):
        html_url = cast(Optional[str], data.get("html_url"))
        draft = cast(bool, data.get("draft", draft))
        id = cast(int, data.get("id", 0))
        name = cast(str, data.get("name"))
        tag_name = cast(str, data.get("tag_name"))

        return ReleaseInfo(url=html_url, is_draft=draft, id=id, ref=name, tag=tag_name)

    def add_release(
        self,
        log: commit.ChangeLog,
        setup: commit.LogSetup,
        git: commit.Git,
        draft: bool,
    ) -> commit.ReleaseInfo:
        if setup.curr_tag is None:
            git.rt.fatal(f"New tag is needed.")

        git.push_with_refs(self.remote, "main")

        body = msg.ReleaseMessage(setup).format_changelog(log)

        rel = {
            "tag_name": setup.curr_tag,
            "name": setup.curr_tag,
            "body": body,
            "draft": draft,
            "prerelease": len(setup.curr_tag.split("-", 1)) > 1,
        }

        flags: List[str] = []
        for name, value in rel.items():
            is_str = isinstance(value, str)
            if not is_str:
                value = json.dumps(value)
            flags.append("-f" if is_str else "-F")
            flags.append(f"{name}={value}")

        data = self.json_from("/releases", *flags, method="POST", default={})
        return self._release_from_json(data, draft)

    def locate_release(self, release_name: str) -> Optional[commit.ReleaseInfo]:
        releases = self.json_from("/releases", default=[], ro_call=True)
        for release in releases:
            if release.get("name") == release_name:
                return self._release_from_json(release)

        return None

    def upload_to_release(
        self,
        release: commit.ReleaseInfo,
        directory: str,
        names: list[str],
    ):
        with cmd.cd(directory):
            return self._run(
                "gh",
                "release",
                "upload",
                release.tag or release.ref or "",
                *names,
                "--clobber",
            )

    def publish(self, release: commit.ReleaseInfo) -> commit.ReleaseInfo:
        if not isinstance(release, ReleaseInfo):
            return commit.ReleaseInfo(is_draft=False)

        release_id = release.id
        data = self.json_from(
            f"/releases/{release_id}",
            "-f",
            "draft=false",
            "-F",
            "make_latest=legacy",
            method="PATCH",
        )
        return self._release_from_json(data)

    @staticmethod
    def from_repo(git: commit.Git, remote: Optional[str] = None):
        info = _find_github(git, remote)
        if info == _NO_GITHUB:
            return None
        return GitHub(info, git.rt)


SSH_PREFIX = "git@github.com:"
HTTPS_PREFIX = "https://github.com/"
GIT_SUFFIX = ".git"
_NO_REPO = "", ""


def _repo_from_gh_cli(git: commit.Git) -> Tuple[str, str]:
    gh = cmd.which("gh")
    if gh is None:
        return _NO_REPO
    proc = git.rt.capture("gh", "repo", "view", "--json", "name,owner", silent=True)
    if proc.returncode != 0:
        return _NO_REPO
    data = json.loads(proc.stdout)
    repo: Optional[str] = data.get("name")
    org: Optional[str] = data.get("owner", {}).get("login")
    if org and repo:
        return org, repo
    return _NO_REPO


def _repo_from_url(url: str):
    for prefix in [SSH_PREFIX, HTTPS_PREFIX]:
        if not url.startswith(prefix):
            continue

        project_path = url[len(prefix) :]
        if project_path.endswith(GIT_SUFFIX):
            project_path = project_path[: -len(GIT_SUFFIX)]

        chunks = project_path.split("/", 1)
        if len(chunks) == 2:
            org, repo = chunks
            return org, repo

    return _NO_REPO


def _github_from_gh(
    gh_repo: Tuple[str, str], remotes: List[commit.Remote], remote_hint: Optional[str]
):
    if gh_repo == _NO_REPO:
        return _NO_GITHUB

    kept_answer = _NO_GITHUB
    for name, _, url in remotes:
        url_repo = _repo_from_url(url)
        if url_repo != gh_repo:
            continue

        candidate = _GitHub(remote=name, owner=url_repo[0], repo=url_repo[1])
        if remote_hint and remote_hint != name:
            kept_answer = candidate
            continue
        return candidate

    return kept_answer


def _github_from_remotes(remotes: List[commit.Remote], remote_hint: Optional[str]):
    kept_answer = _NO_GITHUB
    for name, _, url in remotes:
        url_repo = _repo_from_url(url)
        if url_repo == _NO_REPO:
            continue

        candidate = _GitHub(remote=name, owner=url_repo[0], repo=url_repo[1])
        if remote_hint and remote_hint != name:
            kept_answer = candidate
            continue
        return candidate

    return kept_answer


def _find_github(git: commit.Git, remote_hint: Optional[str]) -> _GitHub:
    gh_repo = _repo_from_gh_cli(git)
    remotes = list(git.remotes(silent=True))

    candidate = _github_from_gh(gh_repo, remotes, remote_hint)

    if candidate != _NO_GITHUB:
        return candidate

    return _github_from_remotes(remotes, remote_hint)



# Copyright (c) 2025 Marcin Zdun
# This code is licensed under MIT license (see LICENSE for details)

"""
The **proj_flow.base** contains low-level tools for higher-level parts of the
library.
"""

import typing

from . import cmd, matrix, plugins, registry, uname

__all__ = ["cmd", "matrix", "plugins", "registry", "uname"]


def path_get(
    structure: typing.Union[dict, list], dotted_path: str, default: typing.Any = None
):
    ctx: typing.Any = structure
    for name in dotted_path.split("."):
        if isinstance(ctx, dict):
            try:
                ctx = ctx[name]
                continue
            except KeyError:
                return default

        if isinstance(ctx, (list, tuple)):
            try:
                index = int(name)
            except ValueError:
                return default
            except IndexError:
                return default

        return default

    return ctx



# Copyright (c) 2026 Marcin Zdun
# This code is licensed under MIT license (see LICENSE for details)

import json
import sys
import typing
from dataclasses import asdict
from pathlib import Path
from pprint import pprint

import chevron

from proj_flow.api import arg, env
from proj_flow.ext.webidl.base.config import (
    CMakeDirs,
    Output,
    TemplateConfig,
    templates_dir,
)
from proj_flow.ext.webidl.cli.updater import update_file_if_needed
from proj_flow.ext.webidl.model.ast import Definitions
from proj_flow.ext.webidl.registry import webidl_visitors


def aslist(items: list):
    return [asdict(item) for item in items]


T = typing.TypeVar("T")
P = typing.TypeVar("P")


def conv(conv: typing.Callable[[T], P], maybe: T | None) -> P | None:
    if maybe is None:
        return None
    return conv(maybe)


@arg.command("webidl", "gen")
def gen(
    config_path: typing.Annotated[
        str, arg.Argument(help="Input configuration", names=["--cfg"], meta="json-file")
    ],
    binary_dir: typing.Annotated[
        str,
        arg.Argument(
            help="Project binary dir", names=["--binary-dir"], meta="project-binary-dir"
        ),
    ],
    rt: env.Runtime,
):
    """Run mustache codegen on WebIDL definitions"""

    config_filename = Path(config_path)
    cmake_dirs = CMakeDirs.from_config_path(
        config_filename, project_binary_dir=Path(binary_dir)
    )

    global_ctx = cmake_dirs.my_defines()
    config_filename = Path(config_path).absolute()
    config = TemplateConfig.load_config(
        config_filename, global_ctx, cmake_dirs.project_source_dir
    )

    global_ctx["version"] = str(config.version)

    pkg_partials: dict[str, str] = {}
    partials_dir = templates_dir / "partials"
    ext = ".mustache"
    for root, _dirs, files in partials_dir.walk():
        for path in [
            root / filename for filename in files if filename[-len(ext) :] == ext
        ]:
            path_id = path.relative_to(partials_dir).with_name(path.stem)
            lang_id = str(path_id.parents[0])
            partial = path_id.as_posix()[len(lang_id) + 1 :]
            partial = f"pkg-{lang_id}:{partial}"
            pkg_partials[partial] = path.read_text(encoding="UTF-8")

            if rt.verbose:
                print(f"-- WebIDL: adding `{partial}' package partial")

    for rule in config.rules:
        names = list(map(lambda path: path.as_posix(), rule.inputs))
        context = {
            **global_ctx,
            "input": Output.filename_context(names[0], cmake_dirs.project_source_dir),
        }

        try:
            idl = Definitions.parse_and_merge(names, config.ext_attrs)
        except FileNotFoundError as e:
            p = Path(e.filename)
            print(f"{e.strerror}: {p.as_posix()}")
            sys.exit(1)
        if isinstance(idl, list):
            for error in idl:
                print(f"{error.path}:{error.error}")
            sys.exit(1)

        for visitor in webidl_visitors.get():
            visitor.on_definitions(idl)

        for output in rule.outputs:
            if not output.mustache_template:
                continue
            write_message: str | None = None
            if not rt.silent:
                fname = (
                    output.output.relative_to(cmake_dirs.project_source_dir).as_posix()
                    if output.output.is_relative_to(cmake_dirs.project_source_dir)
                    else output.output.as_posix()
                )
                write_message = f"writing {fname}"

            types = output.get_type_replacements()
            interfaces, modules_or_includes = idl.order(types)
            output_context = {
                **output.initial_context,
                **context,
                "modules_or_includes": list(sorted(modules_or_includes)),
                "interfaces": aslist(interfaces),
                "enums": aslist(idl.enum),
                "has_modules_or_includes": not not modules_or_includes,
                "has_interfaces": not not interfaces,
                "has_enums": not not idl.enum,
                "output": Output.filename_context(
                    str(output.output), cmake_dirs.project_binary_dir
                ),
            }

            kwargs = {}
            if output.partials:
                kwargs["partials_path"] = output.partials.as_posix()
            text = output.mustache_template.read_text(encoding="UTF-8")
            try:
                text = chevron.render(
                    text, output_context, partials_dict=pkg_partials, **kwargs
                )
            except chevron.ChevronError as e:
                filename = e.filename or output.mustache_template.as_posix()
                line = conv(lambda num: f":{num}", e.lineno) or ""
                col = (conv(lambda num: f":{num}", e.offset) or "") if e.lineno else ""
                print(f"{filename}{line}{col}: {e.msg}")
                if e.text:
                    print(e.text)
                    if e.offset:
                        spaces = "".join(
                            ["\t" if c == "\t" else " " for c in e.text[: e.offset]]
                        )
                        print(f"{spaces}^")
                sys.exit(1)

            update_file_if_needed(output.output, text, write_message)

            dst = output.output.with_name(output.output.name + ".json")
            update_file_if_needed(dst, json.dumps(output_context, indent=2))



# Copyright (c) 2026 Marcin Zdun
# This code is licensed under MIT license (see LICENSE for details)

import concurrent.futures
import os
import shutil
from pathlib import Path
from typing import Any, Generator

from proj_flow.ext.test_runner.driver.test import Env, Test
from proj_flow.ext.test_runner.utils.counters import Counters, TaskResult, color


def _task(
    runtime: Env, tested: Test, current_counter: int
) -> tuple[int, str, str | None, str]:
    env = runtime.with_random_temp_subdir()

    test_counter = f"{color.counter}[{current_counter:>{env.counter_digits}}/{env.counter_total}]{color.reset}"
    test_name = f"{color.name}{tested.name}{color.reset}"
    test_id = f"{test_counter} {test_name}"

    print(test_id)
    os.makedirs(env.tempdir, exist_ok=True)

    result = tested.run(env)
    if result is None:
        return (TaskResult.SKIPPED, test_id, None, env.tempdir)

    actual, files = result
    reports: list[str] = []

    saved = False
    if tested.expected is None:
        tested.data["expected"] = actual.as_dict()
        tested.store()
        saved = True

    for file in files:
        fixed = file.patched(env, tested.cwd, tested.patches)

        if fixed.needs_saving:
            copied = fixed.copy_file()
            saved = saved or copied

        elif fixed.generated.content != fixed.template.content:
            reports.append(tested.report_file(fixed))

    if saved:
        return (TaskResult.SAVED, test_id, None, env.tempdir)

    clipped = tested.clip(actual)

    if isinstance(clipped, str):
        return (TaskResult.CLIP_FAILED, test_id, clipped, env.tempdir)

    if actual != tested.expected and clipped != tested.expected:
        reports.append(tested.report_io(actual))

    if reports:
        reports.append(tested.test_footer(env, env.tempdir))
        return (TaskResult.FAILED, test_id, "\n".join(reports), env.tempdir)

    return (TaskResult.OK, test_id, None, env.tempdir)


def run_and_report_tests(
    independent_tests: list[tuple[Test, int]],
    linear_tests: list[tuple[Test, int]],
    install_dir: Path,
    env: Env,
    thread_count: int,
):
    counters = Counters()

    if independent_tests:
        _report_tests(counters, _run_async_tests(independent_tests, env, thread_count))

    _report_tests(counters, _run_sync_tests(linear_tests, env))

    shutil.rmtree(install_dir, ignore_errors=True)
    shutil.rmtree("build/.testing", ignore_errors=True)

    if not counters.summary(len(independent_tests) + len(linear_tests)):
        return 1

    return 0


def _run_async_tests(tests: list[tuple[Test, int]], env: Env, thread_count: int):
    with concurrent.futures.ThreadPoolExecutor(thread_count) as executor:
        futures: list[concurrent.futures.Future[tuple[int, str, str | None, str]]] = []
        for test, counter in tests:
            futures.append(executor.submit(_task, env, test, counter))

        for future in concurrent.futures.as_completed(futures):
            yield future.result()


def _run_sync_tests(tests: list[tuple[Test, int]], env: Env):
    for test, counter in tests:
        yield _task(env, test, counter)


def _report_tests(
    counters: Counters,
    results: Generator[tuple[int, str, str | None, str], Any, None],
):
    for outcome, test_id, message, tempdir in results:
        counters.report(outcome, test_id, message)
        shutil.rmtree(tempdir, ignore_errors=True)



# Copyright (c) 2026 Marcin Zdun
# This code is licensed under MIT license (see LICENSE for details)

from dataclasses import dataclass, field


class color:
    reset = "\033[m"
    counter = "\033[2;49;92m"
    name = "\033[0;49;90m"
    failed = "\033[0;49;91m"
    passed = "\033[2;49;92m"
    skipped = "\033[0;49;34m"


class TaskResult:
    OK = 0
    SKIPPED = 1
    SAVED = 2
    FAILED = 3
    CLIP_FAILED = 4


@dataclass
class Counters:
    error_counter: int = 0
    skip_counter: int = 0
    save_counter: int = 0
    echo: list[str] = field(default_factory=list)

    def report(self, outcome: int, test_id: str, message: str | None):
        if outcome == TaskResult.SKIPPED:
            print(f"{test_id} {color.skipped}SKIPPED{color.reset}")
            self.skip_counter += 1
            return

        if outcome == TaskResult.SAVED:
            print(f"{test_id} {color.skipped}saved{color.reset}")
            self.skip_counter += 1
            self.save_counter += 1
            return

        if outcome == TaskResult.CLIP_FAILED:
            msg = f"{test_id} {color.failed}FAILED (unknown check '{message}'){color.reset}"
            print(msg)
            self.echo.append(msg)
            self.error_counter += 1
            return

        if outcome == TaskResult.OK:
            print(f"{test_id} {color.passed}PASSED{color.reset}")
            return

        if message is not None:
            print(message)
        msg = f"{test_id} {color.failed}FAILED{color.reset}"
        print(msg)
        self.echo.append(msg)
        self.error_counter += 1

    def summary(self, counter: int):
        print(f"Failed {self.error_counter}/{counter}")
        if self.skip_counter > 0:
            skip_test = "test" if self.skip_counter == 1 else "tests"
            if self.save_counter > 0:
                print(
                    f"Skipped {self.skip_counter} {skip_test} (including {self.save_counter} due to saving)"
                )
            else:
                print(f"Skipped {self.skip_counter} {skip_test}")

        if len(self.echo):
            print()
        for echo in self.echo:
            print(echo)

        return self.error_counter == 0



# Copyright (c) 2024 Marcin Zdun
# This code is licensed under MIT license (see LICENSE for details)

"""
The **proj_flow.log.rich_text.markdown** provides details of making Markdown
changelogs.
"""

from typing import List

from proj_flow.log import commit
from proj_flow.log.rich_text import api


class FileUpdate(api.FileUpdate):
    def __init__(self, setup: commit.LogSetup, commit_date: str):
        super().__init__(setup, commit_date)

    def intro_lines(self) -> List[str]:
        version = (self.setup.curr_tag or " *unknown*")[1:]
        compare = self.setup.commit_listing_link()
        if compare:
            return [f"## [{version}]({compare}) ({self.commit_date})", ""]
        return [f"## {version} ({self.commit_date})", ""]

    def outro_lines(self, _: List[str]) -> None:
        pass

    def issue_link(self, ref: str):
        url = self.setup.reference_link(ref)
        if url:
            return f"[{ref}]({url})"
        return ref


class ChangelogGenerator(api.ChangelogGenerator):
    @property
    def ext(self):
        return ".md"

    @property
    def split_re(self):
        return r"\n## "

    @property
    def intro_text(self):
        return (
            super().intro_text
            + " See [Conventional Commits](https://www.conventionalcommits.org/en/v1.0.0/) "
            "and [COMMITS.md](COMMITS.md) for commit guidelines."
        )

    @property
    def formatter(self):
        return FileUpdate

    def main_heading(self, caption: str) -> List[str]:
        return [f"# {caption}", ""]

    def para(self, text: str) -> List[str]:
        return [text, ""]



# Copyright (c) 2025 Marcin Zdun
# This code is licensed under MIT license (see LICENSE for details)

"""
The **proj_flow.log** defines tools for generating various changelog views.
"""

from . import commit, error, fmt, hosting, msg, release, rich_text

__all__ = [
    "commit",
    "error",
    "fmt",
    "hosting",
    "msg",
    "release",
    "rich_text",
]



# Copyright (c) 2026 Marcin Zdun
# This code is licensed under MIT license (see LICENSE for details)

import os
import random
import shlex
import shutil
import string
import subprocess
import sys
from difflib import unified_diff
from pathlib import Path
from typing import cast

import yaml

from proj_flow.ext.test_runner.driver.env import Env
from proj_flow.ext.test_runner.driver.file_writes import FileWrite
from proj_flow.ext.test_runner.utils.io import ProcessIO

try:
    from yaml import CDumper as Dumper
    from yaml import CLoader as Loader
except ImportError:
    from yaml import Dumper, Loader

_flds = ["Return code", "Standard out", "Standard err"]
_streams = ["stdout", "stderr"]


def _last_enter(value: str):
    if len(value) and value[-1] == "\n":
        value = value[:-1] + "\\n"
    return value + "\n"


def _diff(expected, actual):
    expected = _last_enter(expected).splitlines(keepends=False)
    actual = _last_enter(actual).splitlines(keepends=False)
    return "\n".join(list(unified_diff(expected, actual, lineterm=""))[2:])


def str_presenter(dumper: Dumper, data: str):
    if "\n" in data:
        return dumper.represent_scalar("tag:yaml.org,2002:str", data, style="|")

    return dumper.represent_scalar("tag:yaml.org,2002:str", data)


yaml.add_representer(str, str_presenter, Dumper=Dumper)


def _test_name(filename: Path) -> str:
    dirname = filename.parent.name
    basename = filename.stem

    def num(s: str):
        items = s.split("-")
        if len(items) < 2:
            return s
        items[0] = f"({items[0]})"
        return " ".join(items)

    return f"{num(dirname)} :: {num(basename)}"


def _paths(key: str, dirs: list[str]):
    vals = [val for val in os.environ.get(key, "").split(os.pathsep) if val != ""]
    named = f"${key}"
    if named in dirs:
        pos = dirs.index(named)
        copy = [*dirs]
        copy[pos:pos] = vals
        vals = copy
    else:
        vals.extend(dirs)
    return os.pathsep.join(vals)


class Test:
    cwd: str
    data: dict
    filename: Path
    name: str
    ok: bool
    post_args: list[list[str]]
    current_env: Env | None
    additional_env: dict[str, str]

    linear: bool
    disabled: bool | str
    lang: str
    args: list[str]
    input: str | None
    post: list[list[str]]
    expected: ProcessIO | None
    check: list[str]
    writes: dict[str, str | dict]
    patches: dict[str, str]
    env: dict[str, str | None]
    prepare: list[list[str]]
    cleanup: list[list[str]]

    def __init__(self, data: dict, filename: Path, count: int):
        self.cwd = os.getcwd()
        self.data = data
        self.filename = filename
        self.name = _test_name(filename)
        self.ok = True
        self.post_args = []
        self.current_env = None
        self.additional_env = {}

        self.linear = cast(bool, data.get("linear", False))
        self.disabled = cast(bool | str, data.get("disabled", False))
        self.lang = cast(str, data.get("lang", "en"))
        self.args = []
        self.post = []
        self.input = None
        self.expected = None
        self.check = ["all"] * len(_streams)
        self.writes = cast(dict[str, str | dict], data.get("writes", {}))
        self.patches = cast(dict[str, str], data.get("patches", {}))
        self.env = {}
        self.prepare = []
        self.cleanup = []

        if isinstance(self.disabled, bool):
            self.ok = not self.disabled
        elif isinstance(self.disabled, str):
            self.ok = self.disabled != sys.platform

        test_input = cast(str | list[str] | None, data.get("input"))
        if isinstance(test_input, list):
            test_input = "\n".join(test_input)
        self.input = test_input

        rebuild = False

        _args = cast(str | list[str] | None, data.get("args"))
        if _args is None:
            self.ok = False
            return
        elif isinstance(_args, list):
            rebuild = True
            data["args"] = shlex.join(_args)
            self.args = _args
        else:
            self.args = shlex.split(_args)

        self.post_args, rebuild = self.__join_args("post", data, rebuild)

        if "expected" not in data:
            self.ok = False
            return

        _expected = cast(None | dict[str, str | list[str]], data.get("expected"))
        if isinstance(_expected, dict):
            returncode = cast(None | int, _expected.get("return-code"))
            stdout = _expected.get("stdout")
            stderr = _expected.get("stderr")
            if not (
                isinstance(returncode, int)
                and isinstance(stdout, (str, list))
                and isinstance(stderr, (str, list))
            ):
                self.ok = False
                return
            if isinstance(stdout, list):
                stdout = "\n".join(stdout)
            if isinstance(stderr, list):
                stderr = "\n".join(stderr)
            self.expected = ProcessIO(
                returncode=returncode, stdout=stdout, stderr=stderr
            )

        _check = cast(dict[str, str], data.get("check", {}))
        for index in range(len(_streams)):
            self.check[index] = _check.get(_streams[index], self.check[index])

        _env = cast(dict[str, str | list[str] | None], data.get("env", {}))
        self.env = {
            key: _paths(key, value) if isinstance(value, list) else value
            for key, value in _env.items()
        }

        self.prepare, rebuild = self.__join_args("prepare", data, rebuild)
        self.cleanup, rebuild = self.__join_args("cleanup", data, rebuild)

        if rebuild:
            if self.expected:
                data["expected"] = self.expected.as_dict()
            self.store()

    def __join_args(self, key: str, data: dict, rebuild: bool):
        _rebuild = rebuild
        result: list[list[str]] = []

        lines = cast(str | list[str | list[str]], data.get(key, []))
        if isinstance(lines, str):
            lines = [lines]

        for index in range(len(lines)):
            cmd = lines[index]
            if isinstance(cmd, str):
                result.append(shlex.split(cmd))
                continue

            _rebuild = True
            data[key][index] = shlex.join(cmd)
            result.append(cmd)

        return result, _rebuild

    def run_cmds(self, env: Env, ops: list[list[str]], tempdir: str) -> bool | None:
        saved = self.current_env
        self.current_env = env
        try:
            for op in ops:
                orig = [*op]
                is_safe = False
                try:
                    name = op[0]
                    if name[:5] == "safe-":
                        name = name[5:]
                        is_safe = True
                    min_args, cb = env.handlers[name]
                    op = op[1:]
                    if len(op) < min_args:
                        return None
                    cb(self, [env.expand(o, tempdir, cwd=self.cwd) for o in op])
                except Exception as ex:
                    if op[0] != "safe-rm":
                        print("Problem while handling", shlex.join(orig))
                        print(ex)
                        raise
                    if is_safe:
                        continue
                    return None
        finally:
            self.current_env = saved
        return True

    def run(self, environment: Env) -> tuple[ProcessIO, list[FileWrite]] | None:
        root = os.path.join(
            "build",
            ".testing",
            "".join(random.choice(string.ascii_letters) for _ in range(16)),
        )
        root = self.cwd = os.path.join(self.cwd, root)
        os.makedirs(root, exist_ok=True)

        prep = self.run_cmds(environment, self.prepare, environment.tempdir)
        if prep is None:
            return None

        expanded = [
            environment.expand(arg, environment.tempdir, self.cwd, self.additional_env)
            for arg in self.args
        ]
        post_expanded = [
            [
                environment.expand(
                    arg, environment.tempdir, self.cwd, self.additional_env
                )
                for arg in cmd
            ]
            for cmd in self.post_args
        ]

        _env = {name: os.environ[name] for name in os.environ}
        _env["LANGUAGE"] = self.lang
        for key in self.env:
            value = self.env[key]
            if value:
                _env[key] = environment.expand(value, environment.tempdir, cwd=self.cwd)
            elif key in _env:
                del _env[key]

        cwd = None if self.linear else self.cwd
        io = ProcessIO()
        kwargs = {}
        if self.input is not None:
            kwargs["input"] = self.input.encode()
        proc = subprocess.run(
            [environment.target, *expanded],
            capture_output=True,
            env=_env,
            cwd=self.cwd,
            **kwargs,
        )
        io.append(proc)

        for sub_expanded in post_expanded:
            if io.returncode != 0:
                break
            proc: subprocess.CompletedProcess = subprocess.run(
                [environment.target, *sub_expanded],
                capture_output=True,
                env=_env,
                cwd=cwd,
            )
            io.append(proc)

        expected_files: list[FileWrite] = []
        for key, value in self.writes.items():
            if isinstance(value, str):
                expected_files.append(
                    FileWrite.load(key, value, environment, cwd=self.cwd, save=False)
                )
            elif isinstance(value, dict):
                path = cast(str | None, value.get("path"))
                save = cast(bool, value.get("save", False))
                if not isinstance(path, str) or not isinstance(save, bool):
                    continue

                expected_files.append(
                    FileWrite.load(key, path, environment, cwd=self.cwd, save=save)
                )

        clean = self.run_cmds(environment, self.cleanup, environment.tempdir)
        if clean is None:
            return None

        return (
            environment.patch_io(io, self.cwd, self.patches),
            expected_files,
        )

    def clip(self, actual: ProcessIO) -> str | ProcessIO:
        if not self.expected:
            return actual

        returncode, streams = actual.returncode, [actual.stdout, actual.stderr]
        expected = [self.expected.stdout, self.expected.stderr]

        for ndx in range(len(self.check)):
            check = self.check[ndx]
            if check != "all":
                ex = expected[ndx]
                ex_len = len(ex)
                if check == "begin":
                    if len(streams[ndx]) > ex_len:
                        streams[ndx] = streams[ndx][:ex_len]
                elif check == "end":
                    streams[ndx] = streams[ndx][-ex_len:]
                else:
                    return check
        return ProcessIO(returncode=returncode, stdout=streams[0], stderr=streams[1])

    @staticmethod
    def text_diff(
        header: str, expected: str, actual: str, pre_mark: str = "", post_mark: str = ""
    ):
        return f"""{header}
  Expected:
    {pre_mark}{repr(expected)}{post_mark}
  Actual:
    {pre_mark}{repr(actual)}{post_mark}

Diff:
{_diff(expected, actual)}
"""

    def report_io(self, actual: ProcessIO):
        result = ""
        if not self.expected:
            return result

        if actual.returncode != self.expected.returncode:
            result += f"""{_flds[0]}
  Expected:
    {repr(self.expected.returncode)}
  Actual:
    {repr(actual.returncode)}
"""
        streams = [
            (_flds[1], self.check[0], actual.stdout, self.expected.stdout),
            (_flds[2], self.check[1], actual.stderr, self.expected.stderr),
        ]
        for header, check, actual_stream, expected_stream in streams:
            if actual_stream == expected_stream:
                continue

            if result:
                result += "\n"

            pre_mark = "..." if check == "end" else ""
            post_mark = "..." if check == "begin" else ""
            result += Test.text_diff(
                header,
                expected_stream,
                actual_stream,
                pre_mark,
                post_mark,
            )

        return result

    def report_file(
        self,
        file: FileWrite,
    ):
        header = (
            f"{file.generated.filename} [{'BIN' if file.generated.binary else 'TXT'}]"
        )

        if not file.generated.content:
            header += "\n  New file was not present to be read"
        if not file.template.content:
            header += (
                f"\n  Test file {file.template.filename} was not present to be read"
            )

        if not file.generated.content or not file.template.content:
            return header

        if file.binary:
            return header + "\n  Binary files differ"

        return Test.text_diff(
            header,
            expected=cast(str, file.template.content),
            actual=cast(str, file.generated.content),
        )

    def test_footer(self, env: Env, tempdir: str):
        _env = {}
        _env["LANGUAGE"] = self.lang
        for key in self.env:
            value = self.env[key]
            if value:
                _env[key] = value
            elif key in _env:
                del _env[key]
        if env.reportable_env_prefix:
            for key in os.environ:
                if key.startswith(env.reportable_env_prefix) and key not in _env:
                    _env[key] = os.environ[key]

        expanded = [env.expand(arg, tempdir, cwd=self.cwd) for arg in self.args]
        call = " ".join(
            shlex.quote(arg.replace(os.sep, "/"))
            for arg in [
                *["{}={}".format(key, val) for key, val in _env.items()],
                env.target,
                *expanded,
            ]
        )
        return f"{call}\ncwd: {self.cwd}\ntest: {self.filename}"

    def nullify(self, lang: str | None):
        if lang is not None:
            self.lang = lang
            self.data["lang"] = lang
        self.expected = None
        self.data["expected"] = None
        self.store()

    def store(self):
        with self.filename.open("wb") as f:
            yaml.dump(
                self.data,
                stream=f,
                Dumper=Dumper,
                width=1024,
                allow_unicode=True,
                default_flow_style=False,
                sort_keys=False,
                encoding="UTF-8",
            )

    def path(self, filename):
        return os.path.join(self.cwd, filename)

    def chdir(self, sub):
        self.cwd = os.path.abspath(os.path.join(self.cwd, sub))
        if self.linear:
            os.chdir(self.cwd)

    def ls(self, sub):
        name = self.path(sub)
        for _, dirnames, filenames in os.walk(name):
            names = sorted(
                [
                    *((name.lower(), f"{name}/") for name in dirnames),
                    *((name.lower(), f"{name}") for name in filenames),
                ]
            )
            dirnames[:] = []
            for _, name in names:
                print(name)

    def rmtree(self, sub):
        shutil.rmtree(self.path(sub))

    def cp(self, src: str, dst: str):
        shutil.copy2(self.path(src), self.path(dst))

    def makedirs(self, sub):
        os.makedirs(self.path(sub), exist_ok=True)

    def store_output(self, name: str, args: list[str]):
        env = self.current_env

        if env is not None and args[0] == env.target_name:
            args[0] = env.target

        proc = subprocess.run(args, shell=False, capture_output=True, cwd=self.cwd)
        self.additional_env[name] = proc.stdout.decode("UTF-8").strip()
        print(f"export {name}={self.additional_env[name]}")

    @staticmethod
    def load(filename: Path, count: int):
        contents = filename.read_text(encoding="UTF-8")
        tree = cast(dict, yaml.load(contents, Loader=Loader))
        return Test(tree, filename, count)



# Copyright (c) 2025 Marcin Zdun
# This code is licensed under MIT license (see LICENSE for details)

"""
The **proj_flow.ext.python.rtdocs** defines RTDocs step (`"RTD"`), which uses
``.readthedocs.yaml`` to build the HTML documentation.
"""

import os
import shutil
import subprocess
import sys
from abc import ABC, abstractmethod
from contextlib import contextmanager
from typing import Any, Callable, Dict, List, Optional, cast

from proj_flow.api import env, step
from proj_flow.base import cmd


@step.register
class RTDocs:
    """Runs the jobs defined by ``.readthedocs.yaml`` to build the docs"""

    @property
    def name(self):
        return "RTD"

    def platform_dependencies(self):
        return ["python -m PyYAML"]

    def is_active(self, config: env.Config, rt: env.Runtime) -> bool:
        return (rt.root / ".readthedocs.yaml").is_file()

    def run(self, config: env.Config, rt: env.Runtime) -> int:
        import venv

        import yaml

        with open(rt.root / ".readthedocs.yaml") as rtd_yaml:
            data = yaml.load(rtd_yaml, Loader=yaml.Loader)

            formats = ["html"]

            build_jobs = cast(
                Dict[str, List[str]], data.get("build", {}).get("jobs", {})
            )

            sphinx_configuration = cast(
                Optional[str], data.get("sphinx", {}).get("configuration")
            )

            python_install = cast(
                List[Dict[str, Any]], data.get("python", {}).get("install", [])
            )

        builder: Optional[Builder] = None

        if sphinx_configuration:
            sphinx_configuration = os.path.join(rt.root, sphinx_configuration)
            builder = Sphinx(sphinx_configuration)

        READTHEDOCS_OUTPUT = (
            builder.READTHEDOCS_OUTPUT
            if builder is not None
            else os.path.join(rt.root, "docs/build")
        )
        os.environ["READTHEDOCS_OUTPUT"] = READTHEDOCS_OUTPUT
        os.environ["READTHEDOCS"] = "True"

        jobs: Dict[str, Callable[[], int]] = {
            "create_environment": lambda: _activate_virtual_env(
                venv, os.path.dirname(READTHEDOCS_OUTPUT)
            ),
        }
        if len(python_install):
            jobs["install"] = lambda: _install(python_install)

        if builder:
            for format in formats:
                jobs[f"build/{format}"] = builder.wrap(format)

        for name in build_jobs:
            if name != "build":
                jobs[name] = lambda: _script(build_jobs[name])
                continue

            build_jobs_build = cast(Dict[str, List[str]], build_jobs["build"])
            for format in formats:
                if format not in build_jobs_build:
                    continue
                jobs[f"build/{format}"] = lambda: _script(build_jobs_build[name])

        for job in _job_listing:
            try:
                impl = jobs[job]
            except KeyError:
                continue
            print(f"-- {job}")
            result = impl()
            if result:
                return 1
        return 0


class Builder(ABC):
    """Base class for any recognized builder in the ``.readthedocs.yaml`` config."""

    @property
    @abstractmethod
    def READTHEDOCS_OUTPUT(self) -> str: ...

    @abstractmethod
    def build(self, target: str) -> int: ...

    def wrap(self, target: str) -> Callable[[], int]:
        return lambda: self.build(target)


class Sphinx(Builder):
    """Builder used, if Sphinx config is found in the config file.

    :param config: filename of a Python script Sphinx should use for the configuration
    """

    @property
    def READTHEDOCS_OUTPUT(self) -> str:
        """A ``build/`` subdirectory placed in the same dir, as ``config`` parameter"""
        return self.output

    def __init__(self, config: str):
        self.config = config
        self.source = os.path.dirname(config)
        self.output = os.path.join(os.path.dirname(self.source), "build")

    def build(self, target: str) -> int:
        """Uses ``spinx-build`` to create the documentation.

        :param target: name of the docs format from YAML config
        :returns: exit code forwarded from the build tool
        """
        builder = "latex" if target == "pdf" else target
        print(shutil.which("sphinx-build"))

        return subprocess.run(
            [
                PYTHON_EXECUTABLE,
                "-m",
                "sphinx.cmd.build",
                "-M",
                builder,
                self.source,
                self.READTHEDOCS_OUTPUT,
            ],
            shell=False,
        ).returncode


PYTHON_EXECUTABLE = sys.executable


def _python(
    *args: str,
    module: Optional[str] = None,
    capture_output: bool = True,
) -> subprocess.CompletedProcess:
    if module is not None:
        return subprocess.run(
            [PYTHON_EXECUTABLE, "-m", module, *args],
            shell=False,
            capture_output=capture_output,
        )
    return subprocess.run(
        [PYTHON_EXECUTABLE, *args], shell=False, capture_output=capture_output
    )


def _pip(*args: str, capture_output: bool = False):
    return _python(*args, module="pip", capture_output=capture_output)


_build_targets = [
    "html",
    "htmlzip",
    "pdf",
    "epub",
]

_job_listing = [
    # "post_checkout",
    # "pre_system_dependencies",
    # "post_system_dependencies",
    "pre_create_environment",
    "create_environment",
    "post_create_environment",
    "pre_install",
    "install",
    "post_install",
    "pre_build",
    *(f"build/{tgt}" for tgt in _build_targets),
    "post_build",
]


def _get_venv_path(root: str):
    bindir = os.path.join(".venv", "bin")
    scripts = os.path.join(".venv", "Scripts")

    if os.path.isdir(os.path.join(root, bindir)):
        return bindir

    if os.path.isdir(os.path.join(root, scripts)):
        return scripts

    return None


def _activate_virtual_env(venv, root: str):
    global PYTHON_EXECUTABLE

    with cmd.cd(root):
        exec_ext = ".exe" if sys.platform == "win32" else ""
        python_exec = f"python{exec_ext}"
        bindir = _get_venv_path(root)
        has_venv = bindir is not None and os.path.isfile(
            os.path.join(bindir, python_exec)
        )

        if not has_venv:
            venv.create(".venv", with_pip=True, upgrade_deps=True)
            bindir = _get_venv_path(root)

        if bindir:
            PATH = f"{os.path.abspath(bindir)}{os.pathsep}{os.environ['PATH']}"
            os.environ["PATH"] = PATH
        PYTHON_EXECUTABLE = shutil.which("python") or sys.executable
    return 0


def _install(deps: List[Dict[str, Any]]):
    for dep in deps:
        try:
            requirements = dep["requirements"]
        except KeyError:
            continue

        result = _pip("install", "-q", "-r", requirements).returncode
        if result:
            return result
    return 0


def _script(calls: List[str]):
    for call in calls:
        result = subprocess.run(call, shell=True).returncode
        if result:
            return result
    return 0



# Copyright (c) 2026 Marcin Zdun
# This code is licensed under MIT license (see LICENSE for details)

import sys
from dataclasses import dataclass
from pathlib import Path
from typing import cast

from proj_flow.ext.test_runner.driver.env import Env


@dataclass
class FileContents:
    filename: str
    path: Path
    content: str | bytes | None

    @property
    def empty(self):
        return self.content is None

    @property
    def binary(self):
        return isinstance(self.content, bytes)

    def patched(self, env: Env, cwd: str, patches: dict[str, str]):
        if not isinstance(self.content, str):
            return self
        patched = env.patch(self.content, cwd, patches)
        return FileContents(filename=self.filename, path=self.path, content=patched)


def load_file_contents(filename: str, environment: Env, cwd: str):
    path = Path(environment.expand(filename, environment.tempdir, cwd=cwd))
    try:
        blob = path.read_bytes()
        try:
            text = blob.decode()
        except UnicodeError:
            text = None

        if text is None:
            return FileContents(filename=filename, path=path, content=blob)
        else:
            if sys.platform == "win32":
                text = text.replace("\r", "")
            return FileContents(
                filename=filename,
                path=path,
                content=text,
            )
    except FileNotFoundError:
        return FileContents(filename=filename, path=path, content=None)


@dataclass
class FileWrite:
    generated: FileContents
    template: FileContents
    save: bool

    @property
    def binary(self):
        return self.generated.binary or self.template.binary

    @property
    def needs_saving(self):
        return self.template.content is None or self.save

    def patched(self, env: Env, cwd: str, patches: dict[str, str]):
        return FileWrite(
            generated=self.generated.patched(env, cwd, patches),
            template=self.template,
            save=self.save,
        )

    def copy_file(self):
        if not self.generated.content:
            return False

        self.template.path.parent.mkdir(parents=True, exist_ok=True)
        self.template.path.write_bytes(cast(str, self.generated.content).encode())
        return True

    @staticmethod
    def load(generated_path: str, template_path: str, env: Env, cwd: str, save: bool):
        generated = load_file_contents(generated_path, env, cwd=cwd)
        template = load_file_contents(template_path, env, cwd=cwd)
        return FileWrite(generated=generated, template=template, save=save)



# Copyright (c) 2025 Marcin Zdun
# This code is licensed under MIT license (see LICENSE for details)

"""
The **proj_flow** contains only ``__version__`` to be updated, nothing more.
This is in an attempt to make this module easy to load initially.
"""

__version__ = "0.24.0"



# Copyright (c) 2026 Marcin Zdun
# This code is licensed under MIT license (see LICENSE for details)

"""
The **proj_flow.ext.webidl** provides ``webidl init``, ``webidl cmake``,
``webidl depfile``and ``webidl gen`` commands, allowing for simple
WebIDL/mustache generation.
"""

from . import cli

__all__ = ["cli"]



# Copyright (c) 2025 Marcin Zdun
# This code is licensed under MIT license (see LICENSE for details)

"""
The **proj_flow.minimal.init** implements ``proj-flow init`` command.
"""

import json
import os
import sys
from typing import Annotated, Optional

import yaml

from proj_flow import dependency, flow
from proj_flow.api import arg, ctx, env, init
from proj_flow.base.name_list import name_list
from proj_flow.minimal import base
from proj_flow.project import api, interact


def _project_types():
    return list(map(lambda proj: proj.id, api.project_type.get()))


def _project_help():
    return (
        "Type of project to create. "
        f"Allowed values are: {name_list(_project_types())}"
    )


_output_group = arg.ExclusiveArgumentGroup(opt=True)


@arg.command("init")
def main(
    project: Annotated[
        str,
        arg.Argument(
            help=_project_help,
            meta="project",
            pos=True,
            choices=_project_types,
        ),
    ],
    path: Annotated[
        Optional[str],
        arg.Argument(
            help="Location of initialized project. "
            "The directory will be created, if it does not exist yet. "
            "Defaults to current directory.",
            pos=True,
            default=".",
        ),
    ],
    non_interactive: Annotated[
        bool,
        arg.FlagArgument(
            help="Selects all the default answers.",
            names=["-y", "--yes"],
            group=_output_group,
        ),
    ],
    store: Annotated[
        Optional[str],
        arg.Argument(
            help="Do not create project, store the context in given file.",
            meta="context-file",
            group=_output_group,
        ),
    ],
    answers: Annotated[
        Optional[str],
        arg.Argument(
            help="Take all the answers from the answer sheet.",
            meta="context-file",
            group=_output_group,
        ),
    ],
    save_context: Annotated[
        bool,
        arg.FlagArgument(help="Save the mustache context as YAML.", names=["--ctx"]),
    ],
    rt: env.Runtime,
):
    """Initialize new project"""

    api.load_common_init_setting_extensions()

    setup = interact.ContextSetup(
        dest_path=path,
        interactive=not non_interactive and answers is None,
        simple=store is not None,
        load=answers,
    )

    context_file = store or ".context.yaml"
    save_context = save_context or store is not None
    if path is not None and os.path.basename(path) == "":
        setup.dest_path = os.path.dirname(path)

    try:
        current_project = api.get_project_type(project)
    except api.ProjectNotFound:
        print(f"proj-flow init: error: project type `{project}` is not known")
        return 1

    if path is not None and store is None:
        os.makedirs(path, exist_ok=True)
        os.chdir(path)
        setup.dest_path = None

    errors = dependency.verify(dependency.gather(init.__steps))
    if len(errors) > 0:
        if not rt.silent:
            for error in errors:
                print(f"proj-flow: {error}", file=sys.stderr)
        return 1

    context = current_project.get_context(setup, rt)
    if not non_interactive and not rt.silent:
        print()

    if save_context and rt.verbose:
        lines = yaml.dump(context, indent=4).rstrip().split("\n")
        for line in lines:
            rt.message("[CONTEXT]", line)

    if save_context and not rt.dry_run:
        with open(context_file, "w", encoding="UTF-8") as jsonf:
            if os.path.splitext(context_file)[1] == ".json":
                json.dump(context, jsonf, indent=4)
            else:
                yaml.dump(context, jsonf, indent=4)

    if store is not None:
        return 0

    flow.layer.copy_license(rt, context)
    if not rt.silent:
        print()

    layers = flow.layer.gather_package_layers(ctx.package_root, context)
    for fs_layer in layers:
        fs_layer.run(rt, context)

    if save_context and not rt.dry_run:
        with open(".gitignore", "ab") as ignoref:
            ignoref.write("\n/.context.yaml\n".encode("UTF-8"))

    current_project.append_extensions(context)

    base.GitInit.layers = layers

    steps = sorted((step.priority(), i, step) for i, step in enumerate(init.__steps))
    for _, _, step in steps:
        step.postprocess(rt, context)



# Copyright (c) 2026 Marcin Zdun
# This file is licensed under MIT license (see LICENSE for details)

from proj_flow.api import arg
from proj_flow.ext.tools.pragma_once import pragma_once
from proj_flow.ext.tools.run_linter import run_linter

__all__ = ["tools", "pragma_once", "run_linter"]


@arg.command("tools")
def tools():
    """Run various C++ tools and helpers"""



# Copyright (c) 2024 Marcin Zdun
# This code is licensed under MIT license (see LICENSE for details)

"""
The **proj_flow.log.fmt** contains formatting routines for various targets.
"""

from proj_flow.log import commit, msg


def format_commit_message(log: commit.ChangeLog):
    return msg.CommitMessage().format_changelog(log)



# Copyright (c) 2024 Marcin Zdun
# This code is licensed under MIT license (see LICENSE for details)

"""
The **proj_flow.log.release** performs a relase on the hosting service.
"""

import typing
from typing import Iterable, Optional, Union

from proj_flow import api
from proj_flow.api import env
from proj_flow.base import registry
from proj_flow.log import commit, fmt, rich_text
from proj_flow.log.error import NoProjectError, VersionNotAdvancing

OneOrMoreStrings = Union[str, Iterable[str]]


class VersionUpdater:

    def on_version_change(
        self, rt: env.Runtime, new_version: str
    ) -> Optional[OneOrMoreStrings]:
        return None

    def on_version_change_tags(
        self, rt: env.Runtime, new_version: str, tags: list[str]
    ) -> Optional[OneOrMoreStrings]:
        return self.on_version_change(rt, new_version)


version_updaters = registry.Registry[VersionUpdater]("VersionUpdater")


def _bump_version(ver: str, level: commit.Level):
    split = ver.split("-", 1)
    if len(split) == 2:
        stability = f"-{split[1]}"
    else:
        stability = ""

    semver = [int(s) for s in split[0].split(".")]
    while len(semver) < 3:
        semver.append(0)
    semver = semver[:3]

    if level.value > commit.Level.BENIGN.value:
        # This turns [1, 2, 3] through 4 - x into [3, 2, 1]
        lvl = commit.Level.BREAKING.value - level.value
        semver[lvl] += 1
        for index in range(lvl + 1, len(semver)):
            semver[index] = 0

    return ".".join(str(v) for v in semver) + stability


def _get_project(rt: env.Runtime):
    def wrap(suite: api.release.ProjectSuite):
        return suite.get_project(rt)

    return wrap


def add_release(
    rt: env.Runtime,
    dbg_changelog: bool,
    forced_level: typing.Optional[commit.Level],
    take_all: bool,
    draft: bool,
    generator: rich_text.api.ChangelogGenerator,
    git: commit.Git,
    hosting: commit.Hosting,
):
    """
    Bumps the project version based on current git logs, creates a "chore"
    commit for the change, attaches an annotated tag with the version number
    and pushes it all to hosting.
    """

    suite, project = api.release.project_suites.find(_get_project(rt))

    if not project or not suite:
        raise NoProjectError()

    tags = git.tag_list(silent=True)

    prev_tag = tags[-1] if len(tags) > 0 else None

    rt.message("Tags:")
    for tag in reversed(tags):
        rt.message("  >", tag, "*" if tag == prev_tag else "")

    setup = commit.LogSetup(hosting, prev_tag, None, take_all=take_all)
    changelog, log_level = git.get_log(setup)

    rt.message("Changelog:")
    for grp, links in changelog.items():
        rt.message("  >", grp)
        for link in links:
            rt.message("     ", str(link.summary))

    project_version = f"{project.version}"
    next_version = _bump_version(project_version, forced_level or log_level)
    setup.curr_tag = f"v{next_version}"

    if next_version == project_version:
        raise VersionNotAdvancing(next_version)

    if setup.curr_tag in tags:
        rt.fatal(f"Tag {setup.curr_tag} already exists.")

    files_to_commit: typing.List[str] = []
    if dbg_changelog or not rt.dry_run:
        generator.update_changelog(changelog, setup, rt)
        files_to_commit.append(generator.filename)

        suite.set_version(rt, next_version)
        version_path = suite.get_version_file_path(rt)
        if version_path:
            files_to_commit.append(version_path)

        for updater in version_updaters.get():
            modified = updater.on_version_change_tags(rt, next_version, tags)
            if modified is None:
                continue
            elif isinstance(modified, str):
                files_to_commit.append(modified)
            else:
                files_to_commit.extend(modified)
    else:
        files_to_commit.append(generator.filename)
        version_path = suite.get_version_file_path(rt)
        if version_path:
            files_to_commit.append(version_path)

    commit_message = f"release {next_version}"
    git.add_files(*files_to_commit)
    git.commit(f"chore: {commit_message}{fmt.format_commit_message(changelog)}")
    git.annotated_tag(setup.curr_tag, commit_message)

    if hosting.is_active:
        info = hosting.add_release(changelog, setup, git, draft)
        if info.url:
            msg = "Visit draft at" if info.is_draft else "Visit release at"
            rt.message(msg, info.url, level=env.Msg.ALWAYS)

    return setup.curr_tag



# Copyright (c) 2025 Marcin Zdun
# This code is licensed under MIT license (see LICENSE for details)

"""
The **proj_flow.api.arg** is used by various commands to declare CLI arguments.
"""

import argparse
import inspect
import typing
from dataclasses import dataclass, field

from proj_flow.base import inspect as _inspect

T = typing.TypeVar("T")
LazyArgument = typing.Union[T, typing.Callable[[], T]]


def _eval(arg: LazyArgument[T]) -> T:
    if callable(arg):
        return typing.cast(T, arg())
    return arg


class _Completable(typing.Protocol):
    completer: _inspect.Function


@dataclass
class ExclusiveArgumentGroup:
    opt: bool = False

    def visit(self, parser: argparse._ActionsContainer) -> argparse._ActionsContainer:
        self.opt
        return parser.add_mutually_exclusive_group(required=not self.opt)


@dataclass
class Argument:
    help: LazyArgument[str] = ""
    pos: bool = False
    names: LazyArgument[typing.List[str]] = field(default_factory=list)
    nargs: LazyArgument[typing.Union[str, int, None]] = None
    opt: LazyArgument[typing.Optional[bool]] = None
    meta: LazyArgument[typing.Optional[str]] = None
    action: LazyArgument[typing.Union[str, argparse.Action, None]] = None
    default: LazyArgument[typing.Optional[typing.Any]] = None
    choices: LazyArgument[typing.Optional[typing.List[str]]] = None
    completer: typing.Optional[_inspect.Function] = None
    group: typing.Optional[ExclusiveArgumentGroup] = None

    def visit(self, parser: argparse._ActionsContainer, name: str):
        kwargs = {}

        self_help = _eval(self.help)
        self_names = _eval(self.names)
        self_nargs = _eval(self.nargs)
        self_opt = _eval(self.opt)
        self_meta = _eval(self.meta)
        self_action = _eval(self.action)
        self_default = _eval(self.default)
        self_choices = _eval(self.choices)

        if self_help is not None:
            kwargs["help"] = self_help
        if self_nargs is not None:
            kwargs["nargs"] = self_nargs
        if self_meta is not None:
            kwargs["metavar"] = self_meta
        if self_default is not None:
            kwargs["default"] = self_default
        if self_action is not None:
            kwargs["action"] = self_action
        if self_choices is not None:
            kwargs["choices"] = self_choices

        names = (
            [name] if self.pos else self_names if len(self_names) > 0 else [f"--{name}"]
        )

        if self.pos:
            kwargs["nargs"] = "?" if self_opt else 1
        else:
            kwargs["dest"] = name
            kwargs["required"] = not self_opt

        action = parser.add_argument(*names, **kwargs)
        if self.completer:
            typing.cast(_Completable, action).completer = self.completer

        return action


class FlagArgument(Argument):
    def __init__(
        self,
        help: str = "",
        names: typing.List[str] = [],
        group: typing.Optional[ExclusiveArgumentGroup] = None,
    ):
        super().__init__(
            help=help,
            names=names,
            group=group,
            opt=True,
            action="store_true",
            default=False,
        )


@dataclass
class _Command:
    name: str
    entry: typing.Optional[_inspect.Function]
    doc: typing.Optional[str]
    subs: typing.Dict[str, "_Command"]

    def add(
        self,
        names: typing.List[str],
        entry: _inspect.Function,
        doc: typing.Optional[str],
    ):
        name = names[0]
        rest = names[1:]
        if len(rest):
            try:
                child = self.subs[name]
            except KeyError:
                child = _Command(name, None, None, {})
                self.subs[name] = child

            child.add(rest, entry, doc)
            return

        try:
            child = self.subs[name]
            child.entry = entry
            child.doc = doc
        except KeyError:
            self.subs[name] = _Command(name, entry, doc, {})


_known_commands = _Command("", None, None, {})
_autodoc = {
    "proj_flow.flow.configs.Configs": "Current configuration list.",
    "proj_flow.api.env.Runtime": "Tools and print messages, while respecting ``--dry-run``, ``--silent`` and ``--verbose``.",
    "proj_flow.cli.argument.Command": "The Command object attached to this @command function.",
}


def command(*name: str):
    def wrap(function: object):
        entry = typing.cast(_inspect.Function, function)
        global _known_commands
        orig_doc = inspect.getdoc(entry)
        _known_commands.add(list(name), entry, orig_doc)

        doc = orig_doc or ""
        if doc:
            doc += "\n\n"
        doc += f":call: ``proj-flow {' '.join(name)}``\n\n"
        doc += f":call: ``./flow {' '.join(name)}``\n\n"

        for arg in _inspect.signature(entry):
            help = ""
            for meta in arg.metadata:
                if isinstance(meta, Argument):
                    help = meta.help
                    if help:
                        break

            if not help:
                full_name = f"{arg.type.__module__}.{arg.type.__name__}"
                help = _autodoc.get(full_name, "")

            doc += f":param {_inspect.type_name(arg.type)} {arg.name}: {help}\n"

        entry.__doc__ = doc

        return entry

    return wrap


def get_commands():
    return _known_commands



# Copyright (c) 2026 Marcin Zdun
# This code is licensed under MIT license (see LICENSE for details)

import shlex
import subprocess
import urllib.parse
from dataclasses import dataclass
from pathlib import Path
from typing import Annotated, Any, cast

import requests_cache
from yaml import load, safe_load

from proj_flow.api.arg import Argument, command
from proj_flow.api.env import Runtime

try:
    from yaml import CLoader as Loader
except ImportError:
    from yaml import Loader

IMAGE = "github/super-linter"
PREFIX = f"{IMAGE}/"


@dataclass
class Image:
    repo: str
    name: str
    version: str

    @property
    def full_name(self):
        return f"{self.repo}/{self.name}"

    @property
    def label(self):
        return f"{self.repo}/{self.name}:{self.version}"

    def super_linter_args(self, env: dict[str, str], log_level: str):
        _env = env.copy()
        _env["RUN_LOCAL"] = "true"
        _env["LOG_LEVEL"] = log_level
        _env["USE_FIND_ALGORITHM"] = "true"
        _env["VALIDATE_ALL_CODEBASE"] = "true"

        args = [
            arg
            for pair in (["-e", f"{key}={_env[key]}"] for key in _env)
            for arg in pair
        ]
        return [
            "docker",
            "run",
            "--rm",
            *args,
            "-v" f"{Path().absolute().as_posix()}:/tmp/lint",
            self.label,
        ]


@dataclass
class Step:
    name: str
    image: Image
    env: dict[str, str]


@dataclass
class Job:
    name: str
    steps: list[Step]


@command("tools", "run-linter")
def run_linter(
    log_level: Annotated[
        str,
        Argument(
            help="Select one of log levels (defaults to INFO)",
            choices=["ERROR", "WARN", "NOTICE", "INFO", "DEBUG"],
            default="INFO",
            opt=True,
        ),
    ],
    rt: Runtime,
):
    """Run dockerized super-linter with all the settings from local GitHub workflows"""

    WORKFLOWS = Path(".github/workflows")
    if not WORKFLOWS.is_dir():
        if not rt.silent:
            print("This project has no GitHub workflows")
            return 0

    db_path = Path("build")
    db_path.mkdir(parents=True, exist_ok=True)
    db_path = db_path / "cache_db"

    session = requests_cache.CachedSession(db_path)
    JOBS: list[Job] = []
    for root, dirs, files in WORKFLOWS.walk():
        for filename in files:
            path = root / filename
            if path.suffix not in [".yml", ".yaml"]:
                continue

            with path.open(encoding="UTF-8") as f:
                data = load(f, Loader=Loader)
                try:
                    yml_name = cast(str, data["name"])
                except KeyError:
                    yml_name = path.stem
                JOBS.extend(_get_linter_actions(data, yml_name, session, rt.verbose))

    for job in JOBS:
        job_name, steps = job.name, job.steps
        print(f"job:  \033[1;49;95m{job_name}\033[m")
        for step in steps:
            print(
                f"step: \033[1;49;96m{step.name}\033[m \033[2;49;32m({step.image.full_name}:\033[0;49;32m{step.image.version}\033[2;49;32m)\033[m"
            )
            args = step.image.super_linter_args(step.env, log_level=log_level)
            print("\033[0;49;30m{}\033[m".format(shlex.join(args)))
            subprocess.run(args, shell=False)
            print()


def _download_docker_image_id(
    action: str,
    version: str | None,
    session: requests_cache.CachedSession,
    verbose: bool,
):
    if action != IMAGE and not action.startswith(PREFIX):
        return None

    repo = "https://raw.githubusercontent.com/github/super-linter"
    refs = f"refs/tags/{version}" if version else "refs/heads/main"
    filename = "action.yml"
    if action.startswith(PREFIX):
        filename = f"{action[len(PREFIX):]}/{filename}"

    url = f"{repo}/{refs}/{filename}"
    if verbose:
        print("-- run-linter: Downloading", url)
    response = session.get(url)
    response.raise_for_status()
    data = cast(dict[str, dict[str, str]], safe_load(response.text))
    runs = data.get("runs", {})
    if runs.get("using") != "docker":
        return None

    image = runs.get("image")
    if not image:
        return None

    url = urllib.parse.urlparse(image)
    name, version = url.path[1:].split(":", 1)
    return Image(url.netloc, name, version)


def _get_linter_action(
    step: dict, session: requests_cache.CachedSession, verbose: bool
):
    try:
        id_parts = cast(str, step["uses"]).split("@", 1)
    except KeyError:
        return

    action = id_parts[0]
    action_version = id_parts[1] if len(id_parts) > 1 else None
    image = _download_docker_image_id(action, action_version, session, verbose)
    if not image:
        return

    try:
        name = cast(str, step["name"])
    except KeyError:
        name = cast(str, step["uses"])

    if verbose:
        print(f'-- run-linter: Found {image.label} used by "{name}"')

    try:
        env = cast(dict[str, Any], step["env"])
    except KeyError:
        env = {}

    _env: dict[str, str] = {}
    for key in env:
        value = env[key]
        if isinstance(value, bool):
            _env[key] = "true" if value else "false"
            continue
        if isinstance(value, str) and "${{" in value:
            continue
        _env[key] = value

    return Step(name=name, image=image, env=_env)


def _get_linter_actions(
    data: dict, yml_name: str, session: requests_cache.CachedSession, verbose: bool
):
    result: list[Job] = []
    for job_id, job in cast(dict[str, dict], data["jobs"]).items():
        steps: list[Step] = []
        for step_def in cast(list[dict], job["steps"]):
            step = _get_linter_action(step_def, session, verbose)
            if step is None:
                continue
            steps.append(step)

        if not steps:
            continue

        try:
            name = cast(str, job["name"])
        except KeyError:
            name = job_id

        if name != yml_name:
            name = f"{yml_name} » {name}"

        result.append(Job(name=name, steps=steps))

    return result



# Copyright (c) 2025 Marcin Zdun
# This code is licensed under MIT license (see LICENSE for details)

"""
The **proj_flow.project** contains the inner workings of ``proj-flow init``
command.
"""

from . import api, cplusplus, data, interact

__all__ = ["api", "cplusplus", "data", "interact"]



# Copyright (c) 2024 Marcin Zdun
# This code is licensed under MIT license (see LICENSE for details)

"""
The **proj_flow.log.commit** allows analysing the git commits for
changelog generation.
"""

import re
import secrets
import string
import time
from abc import ABC, abstractmethod
from dataclasses import dataclass, field
from enum import Enum
from typing import Dict, List, NamedTuple, Optional, Tuple

from proj_flow.api import env
from proj_flow.base import registry

COMMIT_SEP = "--{}".format(
    "".join(secrets.choice(string.ascii_letters + string.digits) for i in range(20))
)


def set_commit_sep(sep: str):
    global COMMIT_SEP
    COMMIT_SEP = sep


class Section(NamedTuple):
    key: str
    header: str


class Level(Enum):
    BENIGN = 0
    PATCH = 2
    FEATURE = 3
    BREAKING = 4


FORCED_LEVEL: Dict[str, Level] = {
    "patch": Level.PATCH,
    "fix": Level.PATCH,
    "minor": Level.FEATURE,
    "feat": Level.FEATURE,
    "feature": Level.FEATURE,
    "major": Level.BREAKING,
    "breaking": Level.BREAKING,
    "release": Level.BREAKING,
}

BREAKING_CHANGE = "BREAKING_CHANGE"
TYPES = [
    Section(BREAKING_CHANGE, "Breaking"),
    Section("feat", "New Features"),
    Section("fix", "Bug Fixes"),
]
KNOWN_TYPES = [section.key for section in TYPES]
TYPE_FIX = {"docs": "fix"}
ALL_TYPES = {
    "assets": "Assets",
    "build": "Build System",
    "chore": "Chores",
    "ci": "Continuous Integration",
    "perf": "Performance Improvements",
    "refactor": "Code Refactoring",
    "revert": "Reverts",
    "style": "Code Style",
    "test": "Tests",
}
ISSUE_LINKS = {"refs": "references", "closes": "closes", "fixes": "fixes"}


class Link(NamedTuple):
    scope: str
    summary: str
    hash: str
    short_hash: str
    is_breaking: bool
    breaking_message: List[str]
    references: Dict[str, List[str]]


class Commit(NamedTuple):
    type: str
    link: Link


ChangeLog = Dict[str, List[Link]]


def read_tag_date(tag: str, rt: env.Runtime):
    proc = rt.capture("git", "log", "-n1", "--format=%aI", tag)
    if proc.returncode != 0:
        return time.strftime("%Y-%m-%d")
    return proc.stdout.split("T", 1)[0]


def _get_commit(hash: str, short_hash: str, message: str) -> Optional[Commit]:
    subject, body = (message + "\n\n").split("\n\n", 1)
    split = subject.split(": ", 1)
    if len(split) != 2:
        return None

    encoded, summary = split
    encoded = encoded.strip()
    is_breaking = len(encoded) > 0 and encoded[-1] == "!"
    if is_breaking:
        encoded = encoded[:-1].rstrip()
    type_scope = encoded.split("(", 1)
    if not len(type_scope[0]):
        return None
    scope = ""
    if len(type_scope) == 2:
        scope = ")".join(type_scope[1].split(")")[:-1]).strip()

    references = {}
    lines = body.rstrip().split("\n")
    for indexPlus1 in range(len(lines), 0, -1):
        index = indexPlus1 - 1
        footerLine = lines[index].strip()
        if footerLine == "":
            lines = lines[:-1]
            continue
        footer = footerLine.split(": ", 1)
        if len(footer) == 1:
            break
        lines = lines[:-1]
        name = footer[0].strip().lower()
        if name in ISSUE_LINKS:
            items = [v.strip() for v in footer[1].split(",")]
            key = ISSUE_LINKS[name]
            if key not in references:
                references[key] = []
            references[key] = items + references[key]
            continue

    breaking_change = None
    body = "\n".join(lines).strip().split("BREAKING CHANGE", 1)
    if (
        len(body) > 1
        and (len(body[0]) == 0 or body[0][-1:] == "\n")
        and (len(body[1]) == 1 or body[1][:1] == ":")
    ):
        body = body[1].lstrip(":").strip()
        breaking_change = [
            re.sub(r"\s+", " ", para.strip()) for para in body.split("\n\n")
        ]
        is_breaking = True

    return Commit(
        type_scope[0].strip(),
        Link(
            scope,
            summary,
            hash,
            short_hash,
            is_breaking,
            breaking_change or [],
            references,
        ),
    )


def _level_from_commit(commit: Commit) -> Tuple[Level, str]:
    if commit.link.is_breaking:
        return (Level.BREAKING, commit.link.scope)
    try:
        currentType = TYPE_FIX[commit.type]
        currentScope = commit.type
    except KeyError:
        currentType = commit.type
        currentScope = commit.link.scope
    currentLevel = {"feat": Level.FEATURE, "fix": Level.PATCH}.get(
        currentType, Level.BENIGN
    )
    return (currentLevel, currentScope)


def _sem_ver(tag: str):
    split = tag.split("-", 1)
    if len(split) == 2:
        stability = split[1]
    else:
        stability = ""
    ver = [int(s) for s in split[0].split(".")]
    while len(ver) < 3:
        ver.append(0)
    return [*ver, stability]


@dataclass
class ReleaseInfo:
    url: Optional[str] = None
    is_draft: Optional[bool] = None
    ref: Optional[str] = None
    tag: Optional[str] = None


class Hosting(ABC):
    """
    Generates links to the hosting service.
    """

    #: Project link to hosting service.
    host_link: str

    @property
    def is_active(self):
        """Can publish a release"""
        return True

    def __init__(self, host_link: str):
        self.host_link = host_link

    @abstractmethod
    def single_commit_link(self, link: Link) -> Optional[str]:
        """Link to a single commit on this hosting platform."""
        ...

    @abstractmethod
    def commit_listing_link(self, setup: "LogSetup") -> Optional[str]:
        """
        Link to a comparison page between two tags on this hosting platform.
        """
        ...

    @abstractmethod
    def reference_link(self, ref: str) -> Optional[str]:
        """
        Link to an issue, based on a reference, provided this hosting platform
        can recognize it.
        """
        ...

    @abstractmethod
    def add_release(
        self,
        log: ChangeLog,
        setup: "LogSetup",
        git: "Git",
        draft: bool,
    ) -> ReleaseInfo:
        """
        Publish a release for current setup, putting the log into release
        notes.
        """
        ...

    @abstractmethod
    def locate_release(self, release_name: str) -> Optional[ReleaseInfo]:
        """
        Locate a release by its name.
        """
        ...

    @abstractmethod
    def upload_to_release(
        self,
        release: ReleaseInfo,
        directory: str,
        names: list[str],
    ):
        """
        Upload package artifacts to the release.
        """
        ...

    @abstractmethod
    def publish(self, release: ReleaseInfo) -> ReleaseInfo:
        """
        Publish given release, return updated release info.
        """
        ...


class NoHosting(Hosting):
    """
    Generates links to the hosting service.
    """

    @property
    def is_active(self):
        return False

    def __init__(self):
        super().__init__("")

    def single_commit_link(self, link: Link) -> Optional[str]:
        return None

    def commit_listing_link(self, setup: "LogSetup") -> Optional[str]:
        return None

    def reference_link(self, ref: str) -> Optional[str]:
        return None

    def add_release(
        self, log: ChangeLog, setup: "LogSetup", git: "Git", draft: bool
    ) -> ReleaseInfo:
        return ReleaseInfo(is_draft=False)

    def locate_release(self, release_name: str) -> Optional[ReleaseInfo]:
        return None

    def upload_to_release(
        self,
        release: ReleaseInfo,
        directory: str,
        names: list[str],
    ):
        return None

    def publish(self, release: ReleaseInfo):
        return ReleaseInfo(is_draft=False)


@dataclass
class LogSetup:
    """
    Represents setup for log extarction.
    """

    #: Generate links to the hosting service.
    hosting: Hosting

    #: Start of commit range. If missing, represent the start of current branch.
    prev_tag: Optional[str]

    #: End of commit range. If missing, represent the HEAD of current branch.
    curr_tag: Optional[str]

    #: Allow reacting to commits with typos in scope names
    scope_fix: Dict[str, str] = field(default_factory=dict)

    #: Choose either all Conventional Commits or only ``feat``/``fix`` ones.
    take_all: bool = False

    @property
    def commit_range(self):
        """Commit range for log retrieval."""

        if self.prev_tag is None:
            if self.curr_tag is None:
                return None
            return self.curr_tag

        if self.curr_tag is None:
            return f"{self.prev_tag}..HEAD"

        return f"{self.prev_tag}..{self.curr_tag}"

    def single_commit_link(self, link: Link):
        """Link to a single commit on hosting platform attached to this setup."""
        return self.hosting.single_commit_link(link)

    def commit_listing_link(self):
        """
        Link to a comparison page between two tags on hosting platform
        attached to this setup.
        """
        return self.hosting.commit_listing_link(self)

    def reference_link(self, ref: str):
        """
        Link to an issue, based on a reference, provided hosting platform
        attached to this setup can recognize it.
        """
        return self.hosting.reference_link(ref)


class Remote(NamedTuple):
    name: str
    usage: str
    url: str


class Git:
    rt: env.Runtime

    def __init__(self, rt: env.Runtime):
        self.rt = rt

    def get_log(self, setup: LogSetup, silent=False) -> Tuple[ChangeLog, Level]:
        args = ["git", "log", f"--format=%h %H%n%B%n{COMMIT_SEP}"]

        commit_range = setup.commit_range
        if commit_range:
            args.append(commit_range)

        proc = self.rt.capture(*args, silent=silent)
        return self.parse_log(proc.stdout, COMMIT_SEP, setup)

    def parse_log(self, git_log_output: str, separator: str, setup: LogSetup):
        commitLog: List[Commit] = []
        amassed: List[str] = []
        for line in git_log_output.split("\n"):
            if line == separator:
                if len(amassed):
                    short_hash, hash = amassed[0].split(" ")
                    commit = _get_commit(
                        hash, short_hash, "\n".join(amassed[1:]).strip()
                    )
                    amassed = []

                    if commit is None:
                        continue

                    commitLog.append(commit)
                continue
            amassed.append(line)

        changes: ChangeLog = {}
        level = Level.BENIGN

        for commit in commitLog:
            link = commit.link

            # Hide even from --all
            if commit.type == "chore" and link.summary[:8] == "release ":
                continue
            if "(no-log)" in link.summary:
                continue

            current_level, current_scope = _level_from_commit(commit)
            if current_level.value > level.value:
                level = current_level
            current_type = TYPE_FIX.get(commit.type, commit.type)
            hidden = current_type not in KNOWN_TYPES

            if hidden and not link.is_breaking and not setup.take_all:
                continue
            if hidden and link.is_breaking:
                current_type = BREAKING_CHANGE

            current_scope = setup.scope_fix.get(current_scope, current_scope)
            if current_type not in changes:
                changes[current_type] = []
            changes[current_type].append(
                Link(
                    current_scope,
                    link.summary,
                    link.hash,
                    link.short_hash,
                    link.is_breaking,
                    link.breaking_message,
                    link.references,
                )
            )

        return changes, level

    def remotes(self, silent=False):
        remotes = self.rt.capture("git", "remote", "-v", silent=silent).stdout.split(
            "\n"
        )
        for line in remotes:
            split = line.split("\t", 1)
            if len(split) != 2:
                continue

            remote_name, remote_url_use = split
            remote_use = ""
            remote_url = None
            for use in ["fetch", "push"]:
                suffix = f" ({use})"
                if remote_url_use.endswith(suffix):
                    remote_url = remote_url_use[: -len(suffix)].strip()
                    remote_use = use
                    break

            if remote_url is None:
                continue

            yield Remote(remote_name, remote_use, remote_url)

    def tag_list(self, silent=False):
        """Get list of version tags, in increasing order."""
        tags: List[str] = self.rt.capture("git", "tag", silent=silent).stdout.split(
            "\n"
        )
        versions: List[Tuple[Tuple[int, int, int, str], str]] = []
        for tag in tags:
            if tag[:1] != "v":
                continue
            try:
                value = _sem_ver(tag[1:])
                if value[3] == "":
                    value[3] = "z"
                versions.append(((*value,), tag))
            except Exception:
                continue

        versions.sort()
        return list(map(lambda pair: pair[1], versions))

    def current_branch(self):
        """Get currently checked-out branch"""
        return self.rt.capture("git", "branch", "--show-current").stdout.strip()

    def cmd(self, *command: str):
        return self.rt.cmd("git", *command)

    def add_files(self, *files: str):
        return self.cmd("add", *files)

    def commit(self, message: str):
        return self.cmd("commit", "-m", message)

    def annotated_tag(self, new_tag: str, message: str):
        return self.cmd("tag", "-am", message, new_tag)

    def push_with_refs(self, remote: str, branch: str):
        return self.cmd("push", remote, branch, "--follow-tags", "--force-with-lease")


class HostingFactory(ABC):
    @abstractmethod
    def from_repo(
        self, git: Git, remote: Optional[str] = None
    ) -> Optional[Hosting]: ...


hosting_factories = registry.Registry[HostingFactory]("HostingFactory")



# Copyright (c) 2025 Marcin Zdun
# This code is licensed under MIT license (see LICENSE for details)

"""
The **proj_flow.api.release** provides :class:`ProjectSuite` extension point.
"""

import os
import re
from abc import ABC, abstractmethod
from dataclasses import dataclass
from typing import NamedTuple, Optional, Union

from proj_flow.api import env
from proj_flow.base import registry


class Arg(NamedTuple):
    value: str
    offset: int


NO_ARG = Arg("", -1)


class Decl(NamedTuple):
    name: str
    value: str
    offset: int

    def __str__(self):
        return self.value

    def asArg(self):
        return Arg(self.value, self.offset)


class Version(NamedTuple):
    core: Arg
    stability: Arg

    def __str__(self):
        return f"{self.core.value}{self.stability.value}"


@dataclass
class Project:
    package_root: str
    version: Version

    @property
    def archive_name(self):
        return f"{self.package_root}-{self.version}"

    @property
    def tag_name(self):
        return f"v{self.version}"

    @property
    def package_prefix(self):
        return f"{self.archive_name}-"

    @property
    def package_suffix(self):
        return ""


class ProjectSuite(ABC):
    @abstractmethod
    def get_project(self, rt: env.Runtime) -> Optional[Project]: ...

    def set_version(self, rt: env.Runtime, version: str):
        core = re.split(r"([0-9]+\.[0-9]+\.[0-9]+)", version, maxsplit=1)[1]
        stability = version[len(core) :]

        project = self.get_project(rt)
        if project:
            self.patch_project(rt, project.version.core, core)

        project = self.get_project(rt)
        if project:
            version_pos = project.version

            if len(stability):
                self.patch_project(rt, version_pos.stability, stability)
            elif len(version_pos.stability.value):
                self.patch_project(rt, version_pos.stability, "")

    @abstractmethod
    def get_version_file_path(self, rt: env.Runtime) -> Optional[str]: ...

    def patch_project(self, rt: env.Runtime, pos: Arg, newValue: str):
        path = self.get_version_file_path(rt)

        if not path or not os.path.isfile(path):
            return

        with open(path, "r", encoding="UTF-8") as input:
            text = input.read()

        patched = text[: pos.offset] + newValue + text[pos.offset + len(pos.value) :]

        with open(path, "w", encoding="UTF-8") as input:
            input.write(patched)


project_suites = registry.Registry[ProjectSuite]("ProjectSuite")


def get_project(rt: env.Runtime):
    def wrap(suite: ProjectSuite):
        return suite.get_project(rt)

    _, project = project_suites.find(wrap)
    if project is None:
        rt.fatal(f"Cannot get project information from {rt.root.as_posix()}")
    return project



# Copyright (c) 2025 Marcin Zdun
# This code is licensed under MIT license (see LICENSE for details)

"""
The **proj_flow.minimal.ext.bug_report** adds next version to bug_report.yaml
next to CHANGELOG.rst.
"""

import re
import sys

from proj_flow.api import env
from proj_flow.log import release

YAML_PATH = ".github/ISSUE_TEMPLATE/bug_report.yaml"


def _version(ver: str) -> tuple[int, int, int, str]:
    if ver[:1] == "v":
        ver = ver[1:]

    m = re.match(r"([0-9]+)\.([0-9]+)\.([0-9]+)(.*)", ver)
    if m is None:
        print(f"fatal error: cannot parse {ver}", file=sys.stderr)
        raise SystemExit(1)
    return (int(m.group(1)), int(m.group(2)), int(m.group(3)), ver)


def _prev_version(new_version: str, tags: list[str]):
    current = _version(new_version)
    versions = [_version(tag) for tag in reversed(tags)]
    index = 0
    while index < len(versions):
        comp = versions[index]
        if comp[0] != current[0] or comp[1] != current[1]:
            break

        index += 1
    if index >= len(versions):
        index = 0

    return None if index > len(versions) else versions[index][-1]


@release.version_updaters.add
class VersionUpdater(release.VersionUpdater):
    def on_version_change_tags(
        self, rt: env.Runtime, new_version: str, tags: list[str]
    ):
        old_version = _prev_version(new_version, tags)

        range = [f"      - Current (v{new_version})\n"]
        if old_version:
            range.extend(
                [
                    f"      - Previous (v{old_version})\n",
                    f"      - Older, than v{old_version}\n",
                ]
            )

        with open(YAML_PATH, encoding="UTF-8") as inf:
            lines = inf.readlines()

        try:
            id_index = lines.index("  id: version\n")
            option_index = lines.index("    options:\n", id_index) + 1
            option_index_end = option_index
            while option_index_end < len(lines) and lines[option_index_end].startswith(
                "      - "
            ):
                option_index_end += 1
            lines[option_index:option_index_end] = range
        except ValueError as e:
            print(e, file=sys.stderr)

        with open(YAML_PATH, "w", encoding="UTF-8") as outf:
            outf.writelines(lines)

        return YAML_PATH



# Copyright (c) 2025 Marcin Zdun
# This code is licensed under MIT license (see LICENSE for details)

"""
The **proj_flow.ext.sign.win32** provides code signing with SignTool
from Windows SDKs.
"""

import base64
import json
import os
import platform
import struct
import subprocess
import sys
from typing import Iterable, List, Optional, Tuple

from proj_flow.api.env import Config, Msg, Runtime

from .api import ENV_KEY, SigningTool, get_key, signing_tool

if sys.platform == "win32":
    import winreg

    @signing_tool.add
    class Win32SigningTool(SigningTool):
        def is_active(self, config: Config, rt: Runtime):
            return _is_active(config.os, rt)

        def sign(self, config: Config, rt: Runtime, files: List[str]):
            rt.print("signtool", *(os.path.basename(file) for file in files))
            return _sign(files, rt)

        def is_signable(self, filename: str, as_package: bool):
            if as_package:
                _, ext = os.path.splitext(filename)
                if ext.lower() == ".msi":
                    return True
                # suport NSIS by checking if the archive is a PE
            return _is_pe_exec(filename)

    Version = Tuple[int, int, int]

    machine = {"ARM64": "arm64", "AMD64": "x64", "X86": "x86"}.get(
        platform.machine(), "x86"
    )

    def _find_sign_tool(rt: Runtime) -> Optional[str]:
        with winreg.OpenKeyEx(  # type: ignore
            winreg.HKEY_LOCAL_MACHINE, r"SOFTWARE\Microsoft\Windows Kits\Installed Roots"  # type: ignore
        ) as kits:
            try:
                kits_root = winreg.QueryValueEx(kits, "KitsRoot10")[0]  # type: ignore
            except FileNotFoundError:
                rt.message("sign/win32: No KitsRoot10 value")
                return None

            versions: List[Tuple[Version, str]] = []
            try:
                index = 0
                while True:
                    ver_str = winreg.EnumKey(kits, index)  # type: ignore
                    ver = tuple(int(chunk) for chunk in ver_str.split("."))
                    index += 1
                    versions.append((ver, ver_str))  # type: ignore
            except OSError:
                pass
        versions.sort()
        versions.reverse()
        rt.message(
            "sign/win32: Regarding versions:",
            ", ".join(version[1] for version in versions),
        )
        for _, version in versions:
            # C:\Program Files (x86)\Windows Kits\10\bin\10.0.22621.0\x64\signtool.exe
            sign_tool = os.path.join(kits_root, "bin", version, machine, "signtool.exe")
            if os.path.isfile(sign_tool):
                rt.message("sign/win32: using:", sign_tool)
                return sign_tool
        return None

    def _is_active(os_name: str, rt: Runtime):
        if os_name != "windows":
            return False
        key = get_key(rt)
        return (
            key is not None
            and key.token is not None
            and key.secret is not None
            and _find_sign_tool(rt) is not None
        )

    _IMAGE_DOS_HEADER = "HHHHHHHHHHHHHH8sHH20sI"
    _IMAGE_NT_HEADERS_Signature = "H"
    _IMAGE_DOS_HEADER_size = struct.calcsize(_IMAGE_DOS_HEADER)
    _IMAGE_NT_HEADERS_Signature_size = struct.calcsize(_IMAGE_NT_HEADERS_Signature)
    _MZ = 23117
    _PE = 17744

    def _is_pe_exec(path: str):
        with open(path, "rb") as exe:
            mz_header = exe.read(_IMAGE_DOS_HEADER_size)
            dos_header = struct.unpack(_IMAGE_DOS_HEADER, mz_header)
            if dos_header[0] != _MZ:
                return False

            PE_offset = dos_header[-1]
            if PE_offset < _IMAGE_DOS_HEADER_size:
                return False

            if PE_offset > _IMAGE_DOS_HEADER_size:
                exe.read(PE_offset - _IMAGE_DOS_HEADER_size)

            pe_header = exe.read(_IMAGE_NT_HEADERS_Signature_size)
            signature = struct.unpack(_IMAGE_NT_HEADERS_Signature, pe_header)[0]
            return signature == _PE

    def _sign(files: Iterable[str], rt: Runtime):
        key = get_key(rt)

        if key is None or key.token is None or key.secret is None:
            rt.fatal("sign: the key is missing")

        sign_tool = _find_sign_tool(rt)
        if sign_tool is None:
            rt.message("proj-flow: sign: signtool.exe not found", level=Msg.ALWAYS)
            return 0

        with open("temp.pfx", "wb") as pfx:
            pfx.write(key.secret)

        args = [
            sign_tool,
            "sign",
            "/f",
            "temp.pfx",
            "/p",
            key.token,
            "/tr",
            "http://timestamp.digicert.com",
            "/fd",
            "sha256",
            "/td",
            "sha256",
            *files,
        ]

        result = 1
        try:
            result = subprocess.run(args, shell=False).returncode
        finally:
            os.remove("temp.pfx")

        return result



# Copyright (c) 2024 Marcin Zdun
# This code is licensed under MIT license (see LICENSE for details)

"""
The **proj_flow.log.msg** defines ChangelogMessage and implements
the message for release commit.
"""

import re
from typing import Dict, List

from proj_flow.log import commit


class ChangelogMessage:
    def intro_lines(self) -> List[str]:
        return []

    def section_header(self, lines: List[str], header):
        lines.extend([f"### {header}", ""])

    def issue_link(self, ref: str) -> str:
        return ref

    def scope_text(self, scope: str) -> str:
        if len(scope):
            scope = f"{scope}: "
        return scope

    def summary(self, text: str) -> str:
        return text

    def short_hash_link(self, link: commit.Link) -> str:
        return link.short_hash

    def list_item(self, line_markup: str) -> str:
        return f"- {line_markup}"

    def nothing_to_report(self, lines: List[str]):
        lines.extend(["*Nothing to report.*", ""])

    def outro_lines(self, _: List[str]) -> None:
        pass

    def post_process(self, lines: List[str]):
        return "\n".join(lines)

    def link_str(self, link: commit.Link, show_breaking: bool):
        scope = ""
        if len(link.scope):
            scope = link.scope

        if show_breaking:
            if link.is_breaking:
                if len(scope):
                    scope = f"breaking, {scope}"
                else:
                    scope = f"breaking"
        scope = self.scope_text(scope)

        refs = ""
        for refs_name in link.references:
            refs_links = [
                self.issue_link(issue)
                for issue in link.references[refs_name]
                if issue != ""
            ]
            if len(refs_links) > 0:
                refs += f", {refs_name} "
                if len(refs_links) == 1:
                    refs += refs_links[0]
                else:
                    last = refs_links[-1]
                    listed = ", ".join(refs_links[:-1])
                    refs += f"{listed} and {last}"

        summary = self.summary(link.summary)
        return self.list_item(f"{scope}{summary} ({self.short_hash_link(link)}){refs}")

    def show_links(self, links: List[commit.Link], show_breaking: bool) -> List[str]:
        issues: Dict[str, List[str]] = {}
        for link in links:
            scope = link.scope
            if scope not in issues:
                issues[scope] = []
            issues[scope].append(self.link_str(link, show_breaking))
        result = []
        for scope in sorted(issues.keys()):
            result.extend(issues[scope])
        if len(result):
            result.append("")
        return result

    def format_changelog(self, log: commit.ChangeLog) -> str:
        lines = self.intro_lines()
        lines_marker = len(lines)

        breaking: List[str] = []

        for section in commit.TYPES:
            try:
                type_section = log[section.key]
            except KeyError:
                continue

            show_breaking = section.key != commit.BREAKING_CHANGE

            self.section_header(lines, section.header)
            lines.extend(self.show_links(type_section, show_breaking))
            breaking.extend(_find_breaking_notes(type_section))

        for section in sorted(log.keys()):
            if section in commit.KNOWN_TYPES:
                continue
            type_section = log[section]
            try:
                section_header = commit.ALL_TYPES[section]
            except KeyError:
                section_header = section

            self.section_header(lines, section_header)
            lines.extend(self.show_links(type_section, True))
            breaking.extend(_find_breaking_notes(type_section))

        if len(breaking):
            self.section_header(lines, "BREAKING CHANGES")
            lines.extend(breaking)

        if lines_marker == len(lines):
            self.nothing_to_report(lines)

        self.outro_lines(lines)

        return self.post_process(lines)


class CommitMessage(ChangelogMessage):
    def section_header(self, lines: List[str], header):
        lines.extend([f"{header}:", ""])

    def list_item(self, line_markup: str):
        return f" - {line_markup}"

    def post_process(self, lines: List[str]):
        paras = "\n".join(lines).strip().split("\n\n")

        text = "\n\n".join(CommitMessage.wrap_at_78(para) for para in paras)
        if len(text):
            text = f"\n\n{text}"
        return text

    @staticmethod
    def wrap_at_78(para: str):
        if para[:3] == " - ":
            lines = para.split("\n")
            lines = [
                CommitMessage._wrap_at(75, line[3:], " - ", "   ") for line in lines
            ]
            return "\n".join(lines)
        return CommitMessage._wrap_at(78, para, "", "")

    @staticmethod
    def _wrap_at(length: int, para: str, firstLine: str, nextLines: str):
        result = ""
        line = firstLine
        lineIsDirty = False
        words = para.strip().split(" ")
        for word in words:
            wordLen = len(word)
            if wordLen == 0:
                continue

            lineIsDirty = True
            lineLen = len(line)
            space = " " if lineLen > 0 and line[-1] != " " else ""
            resultingLen = lineLen + len(space) + wordLen
            if resultingLen <= length:
                line = f"{line}{space}{word}"
                continue
            result = f"{result}{line}\n"
            line = f"{nextLines}{word}"

        if lineIsDirty:
            result = f"{result}{line}"
        return result


class ReleaseMessage(ChangelogMessage):
    setup: commit.LogSetup

    def __init__(self, setup: commit.LogSetup):
        self.setup = setup

    def scope_text(self, scope: str):
        if len(scope):
            scope = f"**{scope}**: "
        return scope

    def short_hash_link(self, link: commit.Link):
        url = self.setup.single_commit_link(link)
        if not url:
            return link.short_hash
        return f"[{link.short_hash}]({url})"

    def outro_lines(self, lines: List[str]) -> None:
        url = self.setup.commit_listing_link()
        if url:
            lines.append(f"**Full Changelog**: {url}")


def _find_breaking_notes(links: List[commit.Link]) -> List[str]:
    breaking: List[str] = []
    for link in links:
        if link.breaking_message is None:
            continue
        for para in link.breaking_message:
            text = re.sub(r"\s+", " ", para.strip())
            if text != "":
                breaking.append(text + "\n")
    return breaking


def format_commit_message(log: commit.ChangeLog):
    return CommitMessage().format_changelog(log)



# Copyright (c) 2024 Marcin Zdun
# This code is licensed under MIT license (see LICENSE for details)

"""
The **proj_flow.log.rich_text.re_structured_text** provides details of making
reStructuredText changelogs.
"""

from typing import List

from proj_flow.log import commit, msg
from proj_flow.log.rich_text import api


class FileUpdate(api.FileUpdate):
    def __init__(self, setup: commit.LogSetup, commit_date: str):
        super().__init__(setup, commit_date)

    def intro_lines(self) -> List[str]:
        version = (self.setup.curr_tag or " *unknown*")[1:]
        compare = self.setup.commit_listing_link()

        if compare:
            heading = f"`{version} <{compare}>`_ ({self.commit_date})"
        else:
            heading = f"{version}"

        return [
            heading,
            "=" * len(heading),
            "",
        ]

    def section_header(self, lines: List[str], header):
        lines.extend([header, "-" * len(header), ""])

    def summary(self, text: str) -> str:
        return text.replace("`", "``")

    def outro_lines(self, _: List[str]) -> None:
        pass

    def short_hash_link(self, link: commit.Link):
        url = self.setup.single_commit_link(link)
        if not url:
            return link.short_hash
        return f"`{link.short_hash} <{url}>`_"


class ChangelogGenerator(api.ChangelogGenerator):
    @property
    def ext(self):
        return ".rst"

    @property
    def split_re(self):
        return r"\n`[0-9]+\.[0-9]+[^>]+\>`\_ \([^)]+\)\n==+\n"

    @property
    def formatter(self):
        return FileUpdate

    def main_heading(self, caption: str) -> List[str]:
        line = "=" * len(caption)
        return [line, caption, line, ""]

    def para(self, text: str) -> List[str]:
        return [msg.CommitMessage.wrap_at_78(text), ""]



# Copyright (c) 2025 Marcin Zdun
# This code is licensed under MIT license (see LICENSE for details)

"""
The **proj_flow.project.interact** provides initialization context through
user prompts.
"""

import os
from dataclasses import dataclass
from typing import Any, Callable, List, Optional, Union, cast

from prompt_toolkit import prompt as tk_prompt
from prompt_toolkit.completion import WordCompleter
from prompt_toolkit.formatted_text.base import AnyFormattedText
from prompt_toolkit.shortcuts import CompleteStyle
from prompt_toolkit.validation import Validator

from proj_flow.api import ctx, env
from proj_flow.base import plugins


@dataclass
class _Question:
    key: str
    prompt: str
    value: ctx.Values

    @classmethod
    def load_default(
        cls,
        default: ctx.Setting,
        previous: ctx.SettingsType,
        override: Optional[ctx.StrOrBool],
    ):
        value = default.calc_value(previous)
        if override is not None:
            if isinstance(value, (str, bool)) and type(value) == type(override):
                value = override
            elif isinstance(value, list) and isinstance(override, str):
                value = ctx.move_to_front(override, value)
        return cls(default.json_key, default.prompt, value)

    def interact(self, counter: int, size: int) -> ctx.StrOrBool:
        if isinstance(self.value, str):
            return self._get_str(self.value, counter, size)
        if isinstance(self.value, bool):
            return self._get_flag(self.value, counter, size)
        return self._get_list(self.value, counter, size)

    @property
    def ps(self):
        return self.prompt or f'"{self.key}"'

    def _ps(self, default: ctx.Values, counter: int, size: int) -> AnyFormattedText:
        if isinstance(default, str):
            if default == "":
                return f"[{counter}/{size}] {self.ps}: "
            return [
                ("", f"[{counter}/{size}] {self.ps} ["),
                ("bold", default),
                ("", f"]: "),
            ]
        if isinstance(default, bool):
            b = "bold"
            n = ""
            on_true = (b if default else n, "yes")
            on_false = (b if not default else n, "no")
            return [
                ("", f"[{counter}/{size}] {self.ps} ["),
                on_true,
                ("", " / "),
                on_false,
                ("", f"]: "),
            ]
        return [
            ("", f"[{counter}/{size}] {self.ps} ["),
            ("bold", default[0]),
            ("", f"{''.join(f' / {x}' for x in default[1:])}]: "),
        ]

    def _get_str(self, default: str, counter: int, size: int):
        value = tk_prompt(self._ps(default, counter, size))
        if not value:
            value = default
        return value

    def _get_flag(self, default: bool, counter: int, size: int):
        value = self._tk_prompt(
            default, ["yes", "no", "on", "off", "1", "0"], counter, size
        )
        result = default
        if value:
            result = value.lower() in ["1", "on", "y", "yes"]
        return result

    def _get_list(self, defaults: List[str], counter: int, size: int):
        value = self._tk_prompt(defaults, defaults, counter, size)
        if not value:
            value = defaults[0]
        return value

    def _tk_prompt(
        self,
        defaults: Union[bool, List[str]],
        words: List[str],
        counter: int,
        size: int,
    ):
        def valid(word: str):
            return word == "" or word in words

        validator = Validator.from_callable(valid)
        completer = WordCompleter(words)
        return tk_prompt(
            self._ps(defaults, counter, size),
            validator=validator,
            completer=completer,
            complete_while_typing=True,
            complete_style=CompleteStyle.READLINE_LIKE,
        )


def _project_filter(project: Optional[str]):
    if project is None:

        def impl(setting: ctx.Setting):
            return setting.project is None

        return impl

    def impl(setting: ctx.Setting):
        return setting.project is None or setting.project == project

    return impl


def _prompt(
    wanted: Callable[[ctx.Setting], bool], overrides: ctx.SettingsType
) -> ctx.SettingsType:
    settings: ctx.SettingsType = {}

    defaults = [setting for setting in ctx.defaults if wanted(setting)]
    switches = [setting for setting in ctx.switches if wanted(setting)]

    size = len(defaults) + len(switches)
    counter = 1

    for coll in [defaults, switches]:
        for setting in coll:
            loaded = _Question.load_default(
                setting, settings, overrides.get(setting.json_key)
            )
            value = loaded.interact(counter, size)
            settings[loaded.key] = value
            counter += 1

    return settings


def _all_default(wanted: Callable[[ctx.Setting], bool], overrides: ctx.SettingsType):
    """
    Chooses default answers for all details of newly-crated project.

    :returns: Dictionary with default values of all interactive settings
        and switches.
    """

    settings: ctx.SettingsType = {}

    defaults = [setting for setting in ctx.defaults if wanted(setting)]
    switches = [setting for setting in ctx.switches if wanted(setting)]

    for coll in [defaults, switches]:
        for setting in coll:
            if setting.json_key in overrides:
                settings[setting.json_key] = overrides[setting.json_key]
            else:
                value = _get_default(setting, settings)
                settings[setting.json_key] = value

    return settings


def _fixup(settings: ctx.SettingsType, key: str, fixup: str, force=False):
    value = settings.get(key, "")
    if value != "" and not force:
        return

    value = ctx._build_fixup(settings, fixup)
    settings[key] = value


def _get_default(setting: ctx.Setting, settings: ctx.SettingsType):
    value = setting.calc_value(settings)
    if isinstance(value, list):
        return value[0]
    return value


def _fixup_context(settings: ctx.SettingsType, wanted: Callable[[ctx.Setting], bool]):
    defaults = [setting for setting in ctx.defaults if wanted(setting)]
    hidden = [setting for setting in ctx.hidden if wanted(setting)]

    for setting in hidden:
        value = _get_default(setting, settings)
        if isinstance(value, bool) or value != "":
            settings[setting.json_key] = value

    for coll in [defaults, hidden]:
        for setting in coll:
            _fixup(settings, setting.json_key, setting.fix or "", setting.force_fix)

    try:
        del settings["EXT"]
    except KeyError:
        pass

    return _split_keys(settings)


def _split_keys(settings: dict):
    result = {}
    for key in settings:
        path = key.split(".")
        path_ctx = result
        for step in path[:-1]:
            if step not in path_ctx or not isinstance(path_ctx[step], dict):
                path_ctx[step] = {}
            path_ctx = path_ctx[step]
        path_ctx[path[-1]] = settings[key]
    return result


def _flatten_keys(settings: Any, prefix=""):
    if not isinstance(settings, dict):
        yield (prefix, settings)
        return

    for key in settings:
        next = f"{prefix}{key}."
        for name, value in _flatten_keys(settings[key], next):
            yield (cast(str, name), cast(Any, value))


def _flatten_dict(settings: dict):
    result = {}
    for name, value in _flatten_keys(settings):
        result[name[:-1]] = value
    return result


@dataclass
class ContextSetup:
    """
    Holds the setup for current context gathering.
    """

    #: Provides path name for project, in case it should differ from current
    #: directory.
    dest_path: Optional[str]

    #: Selects, if the initialization process is done through prompts, or not.
    interactive: bool

    #: Do no post-processing and key expanding of the resulting context
    simple: bool

    #: If this setup is non-interactive, use contents of this file for
    #: answers. If any given answer is missing, the default answer would be
    #: taken.
    load: Optional[str]


def get_context(setup: ContextSetup, project: Optional[str], rt: env.Runtime):
    """
    Prompts user to provide details of newly-crated project. If `interactive`
    is true, however, this functions skips the prompts and chooses all the
    default answers.

    :param setup: Selects, if the initialization process is done through
        prompts, or not and how to answer any given question.

    :param project: Alows to select questions for any given language.

    :param rt: Runtime with config loaded in.

    :returns: Dictionary with answers to all interactive settings and switches.
    """

    overrides = rt._cfg.get("defaults", {})
    if setup.dest_path is not None:
        overrides["PROJECT.NAME"] = os.path.basename(setup.dest_path)

    if setup.load is not None:
        data = plugins.load_data(setup.load)
        data = _flatten_dict(_split_keys(data))
        for key, value in data.items():
            overrides[key] = value

    wanted = _project_filter(project)
    result = (
        _all_default(wanted, overrides)
        if not setup.interactive
        else _prompt(wanted, overrides)
    )
    if "COPY.YEAR" in result:
        fast = cast(dict, result)
        fast["COPY.YEAR"] = int(fast["COPY.YEAR"])
    if setup.simple:
        return _split_keys(result)
    return _fixup_context(result, wanted)



# Copyright (c) 2025 Marcin Zdun
# This code is licensed under MIT license (see LICENSE for details)
"""
The **proj_flow.ext.github.switches** provides GitHub-related switches for new
projects.
"""

from proj_flow.api import ctx

ctx.register_common_switch("with.github.actions", "Use Github Actions", True)
ctx.register_common_switch(
    "with.github.auto-release",
    "Make a GitHub release automatically on each closed PR",
    False,
)
ctx.register_common_switch(
    "with.github.social", "Use Github ISSUE_TEMPLATE, CONTRIBUTING.md, etc.", True
)

ctx.register_common_init_setting(
    ctx.Setting(
        "with.github.no-auto-release",
        value=lambda settings: not settings.get("with.github.auto-release"),
    ),
    is_hidden=True,
)



# Copyright (c) 2025 Marcin Zdun
# This code is licensed under MIT license (see LICENSE for details)

"""
The **proj_flow.project.api** defines an extension point for project suites.
"""

from abc import ABC, abstractmethod
from typing import Any, List, NamedTuple, Optional

from proj_flow import base
from proj_flow.api import ctx, env
from proj_flow.project import interact


class ProjectType(ABC):
    name: str
    id: str

    def __init__(self, name: str, id: str):
        self.name = name
        self.id = id

    def register_switch(self, key: str, prompt: str, enabled: bool):
        ctx.register_switch(key, prompt, enabled, self.id)

    def register_internal(self, key: str, value: Any):
        ctx.register_internal(key, value)

    def register_init_setting(self, *settings: ctx.Setting, is_hidden=False):
        ctx.register_init_setting(*settings, is_hidden=is_hidden, project=self.id)

    def get_context(self, setup: interact.ContextSetup, rt: env.Runtime):
        return interact.get_context(setup, self.id, rt)

    def append_extensions(self, context: dict):
        extensions = self.get_extension_list(context)
        if len(extensions) == 0:
            return
        text = "".join(f"\n  - {ext}" for ext in extensions)
        text = f"extensions:{text}\n\n"

        with open(".flow/config.yml", encoding="UTF-8") as config_file:
            content = config_file.read()

        with open(".flow/config.yml", "w", encoding="UTF-8") as config_file:
            config_file.write(text)
            config_file.write(content)

    @abstractmethod
    def get_extension_list(self, context: dict) -> List[str]: ...


project_type = base.registry.Registry[ProjectType]("ProjectType")


class ProjectNotFound(Exception):
    name: str

    def __init__(self, name: str):
        self.name = name


def get_project_type(id: str):
    result, _ = project_type.find(lambda proj: proj.id == id)
    if result is None:
        raise ProjectNotFound(id)
    return result


def load_common_init_setting_extensions():
    env.load_extensions(["proj_flow.ext.github.switches"])



# Copyright (c) 2026 Marcin Zdun
# This code is licensed under MIT license (see LICENSE for details)

import os
import random
import re
import string
from dataclasses import dataclass, replace
from typing import Any, Callable

from proj_flow.ext.test_runner.utils.io import ProcessIO


def _alt_sep(input: str, value: str, var: str):
    split = input.split(value)
    first = split[0]
    split = split[1:]
    for index in range(len(split)):
        m = re.match(r"^(\S+)((\n|.)*)$", split[index])
        if m is None:
            continue
        g2 = m.group(2)
        if g2 is None:
            g2 = ""
        split[index] = "{}{}".format(m.group(1).replace(os.sep, "/"), g2)
    return var.join([first, *split])


@dataclass
class Env:
    target: str
    target_name: str
    build_dir: str
    data_dir: str
    inst_dir: str
    tempdir: str
    version: str
    counter_digits: int
    counter_total: int
    handlers: dict[str, tuple[int, Callable[[Any, list[str]], None]]]
    data_dir_alt: str | None = None
    tempdir_alt: str | None = None
    # TODO: installable patches
    builtin_patches: dict[str, str] | None = None
    reportable_env_prefix: str | None = None

    def with_random_temp_subdir(self):
        temp_instance = "".join(random.choice(string.ascii_letters) for _ in range(16))
        tempdir = f"{self.tempdir}/{temp_instance}"
        tempdir_alt = None
        if self.tempdir_alt is not None:
            tempdir_alt = f"{self.tempdir_alt}{os.sep}{temp_instance}"

        return replace(self, tempdir=tempdir, tempdir_alt=tempdir_alt)

    def expand(
        self, input: str, tempdir: str, cwd: str, additional: dict[str, str] = {}
    ):
        input = (
            input.replace("$TMP", tempdir)
            .replace("$CWD", cwd)
            .replace("$DATA", self.data_dir)
            .replace("$INST", self.inst_dir)
            .replace("$VERSION", self.version)
        )
        for key, value in additional.items():
            input = input.replace(f"${key}", value)
        return input

    def patch_io(self, io: ProcessIO, cwd: str, patches: dict[str, str]):
        stdout = io.stdout
        stderr = io.stderr
        if os.name == "nt":
            stdout = stdout.replace("\r\n", "\n")
            stderr = stderr.replace("\r\n", "\n")
        return ProcessIO(
            returncode=io.returncode,
            stdout=self.patch(stdout, cwd, patches),
            stderr=self.patch(stderr, cwd, patches),
        )

    def patch(self, input: str, cwd: str, patches: dict[str, str]):
        input = _alt_sep(input, cwd, "$CWD")
        input = _alt_sep(input, self.tempdir, "$TMP")
        input = _alt_sep(input, self.data_dir, "$DATA")
        input = input.replace(self.version, "$VERSION")

        if self.tempdir_alt is not None:
            input = _alt_sep(input, self.tempdir_alt, "$TMP")
            if self.data_dir_alt:
                input = _alt_sep(input, self.data_dir_alt, "$DATA")

        builtins = self.builtin_patches or {}

        lines = input.split("\n")
        for patch, replacement in builtins.items():
            pattern = re.compile(patch)
            for lineno in range(len(lines)):
                m = pattern.match(lines[lineno])
                if m:
                    lines[lineno] = m.expand(replacement)

        for patch, replacement in patches.items():
            pattern = re.compile(patch)
            for lineno in range(len(lines)):
                m = pattern.match(lines[lineno])
                if m:
                    lines[lineno] = m.expand(replacement)
        return "\n".join(lines)



# Copyright (c) 2025 Marcin Zdun
# This code is licensed under MIT license (see LICENSE for details)

"""
The **proj_flow.minimal.ext.versions** update version schema files from version-updates
"""

from pathlib import Path
from typing import cast

from proj_flow.api import env
from proj_flow.log import release


@release.version_updaters.add
class VersionUpdater(release.VersionUpdater):
    def on_version_change_tags(
        self, rt: env.Runtime, new_version: str, tags: list[str]
    ) -> release.OneOrMoreStrings | None:
        if not tags:
            return None
        prev_tag = tags[-1]
        old_version = prev_tag[1:] if prev_tag.startswith("v") else prev_tag

        files = cast(dict[str, str], rt._cfg.get("version-updates", {}))
        changed_files: list[str] = []

        rt.message("Config")
        for key, value in files.items():
            if "$VERSION" not in value:
                continue

            path = Path(key)
            previous = value.replace("$VERSION", old_version)
            next = value.replace("$VERSION", new_version)
            rt.message(f"  >> {key} :: {previous} -> {next}")

            try:
                content = path.read_text(encoding="UTF-8")
            except FileNotFoundError:
                rt.fatal(f"File not found: {key}")

            new_content = content.replace(previous, next)
            if new_content != content:
                path.write_text(new_content, encoding="UTF-8")
                changed_files.append(key)

        return changed_files



# Copyright (c) 2025 Marcin Zdun
# This code is licensed under MIT license (see LICENSE for details)

"""
The **proj_flow.base.plugins** provide the plugin enumeration helpers.
"""

import json
from pathlib import Path
from typing import cast

import yaml

try:
    from yaml import CLoader as Loader
except ImportError:
    from yaml import Loader


def load_yaml(filename: Path):
    with open(filename) as src:
        return cast(dict, yaml.load(src, Loader=Loader))


def load_json(filename: Path):
    with open(filename) as src:
        return cast(dict, json.load(src))


LOADERS = {
    ".json": load_json,
    ".yml": load_yaml,
    ".yaml": load_yaml,
}


def load_data(filename: Path):
    ext = filename.suffix
    loader = LOADERS.get(ext.lower())
    if loader:
        try:
            return loader(filename)
        except Exception:
            pass

    for new_ext, loader in LOADERS.items():
        new_filename = filename.with_suffix(new_ext)
        try:
            return loader(new_filename)
        except Exception:
            pass

    return {}



# PYTHON_ARGCOMPLETE_OK

# Copyright (c) 2025 Marcin Zdun
# This code is licensed under MIT license (see LICENSE for details)

"""
The **proj_flow.cli** provides command-line entry for the *Project Flow*.
"""

import argparse
import os
import sys

from proj_flow.api import env
from proj_flow.cli import argument, finder
from proj_flow.flow import steps

__all__ = ["argument", "finder", "main"]


def main():
    """Entry point for ``proj-flow`` tool."""
    try:
        __main()
    except KeyboardInterrupt:
        sys.exit(1)


def _change_dir():
    root = argparse.ArgumentParser(
        prog="proj-flow",
        usage="proj-flow [-h] [--version] [-C [dir]] command ...",
        add_help=False,
    )
    root.add_argument("-C", dest="cd", nargs="?")

    args, _ = root.parse_known_args()
    if args.cd:
        os.chdir(args.cd)


def __main():
    _change_dir()

    flow_cfg = env.FlowConfig(root=finder.autocomplete.find_project())
    steps.clean_aliases(flow_cfg)

    parser = argument.build_argparser(flow_cfg)
    finder.autocomplete(parser)
    args = parser.parse_args()
    argument.expand_shortcuts(parser, args)

    raise SystemExit(parser.find_and_run_command(args))



# Copyright (c) 2026 Marcin Zdun
# This code is licensed under MIT license (see LICENSE for details)

import subprocess
from dataclasses import dataclass, field


@dataclass
class ProcessIO:
    returncode: int = field(default=0)
    stdout: str = field(default_factory=str)
    stderr: str = field(default_factory=str)

    def as_dict(self):
        return {
            "return-code": self.returncode,
            "stdout": self.stdout,
            "stderr": self.stderr,
        }

    def append(self, proc: subprocess.CompletedProcess[bytes]):
        self.returncode = proc.returncode
        if len(self.stdout) and len(proc.stdout):
            self.stdout += "\n"
        if len(self.stderr) and len(proc.stderr):
            self.stderr += "\n"
        self.stdout += proc.stdout.decode()
        self.stderr += proc.stderr.decode()



# Copyright (c) 2025 Marcin Zdun
# This code is licensed under MIT license (see LICENSE for details)

"""
The **proj_flow.ext.github.hosting** adds GitHub hosting environment.
"""

from typing import Optional

from proj_flow.log import commit
from proj_flow.log.hosting.github import GitHub


@commit.hosting_factories.add
class Plugin(commit.HostingFactory):
    def from_repo(
        self, git: commit.Git, remote: Optional[str] = None
    ) -> Optional[commit.Hosting]:
        return GitHub.from_repo(git, remote)



# Copyright (c) 2025 Marcin Zdun
# This code is licensed under MIT license (see LICENSE for details)

"""
The **proj_flow.minimal** defines minimal extension package: ``bootstrap``
and ``run`` commands, with basic set of steps.
"""

from . import base, bootstrap, init, list, run, system

__all__ = ["base", "bootstrap", "init", "list", "run", "system"]



# Copyright (c) 2026 Marcin Zdun
# This code is licensed under MIT license (see LICENSE for details)

import os
import platform
import shutil
import subprocess
import sys
import tempfile
from os import PathLike
from pathlib import Path
from typing import Annotated, cast

from yaml import __version__ as yaml_version

try:
    from yaml import CLoader

    yaml_loader = "CLoader"
except ImportError:
    from yaml import Loader

    yaml_loader = "Loader"

from proj_flow import __version__
from proj_flow.api import arg, env, release
from proj_flow.base.cmake_presets import Presets
from proj_flow.ext.test_runner.driver.commands import HANDLERS
from proj_flow.ext.test_runner.driver.test import Env, Test
from proj_flow.ext.test_runner.driver.testbed import run_and_report_tests

RUN_LINEAR = os.environ.get("RUN_LINEAR", 0) != 0


@arg.command("tools", "test-runner")
def test_runner(
    preset_name: Annotated[
        str,
        arg.Argument(
            help="Set name of CMake build preset",
            meta="CONFIG",
            names=["--preset"],
        ),
    ],
    tests: Annotated[
        str,
        arg.Argument(
            help="Point to directory with the JSON test cases; test cases are enumerated recursively",
            meta="DIR",
        ),
    ],
    version: Annotated[
        str | None,
        arg.Argument(
            help="Select version to patch output with; defaults to automatic detection",
            meta="SEMVER",
            opt=True,
        ),
    ],
    run: Annotated[
        list[str],
        arg.Argument(
            help="Filter the tests to run",
            meta="ID",
            action="extend",
            opt=True,
            nargs="*",
            default=[],
        ),
    ],
    nullify: Annotated[
        bool,
        arg.FlagArgument(
            help='Set the "expected" field of the test cases to null',
        ),
    ],
    rt: env.Runtime,
) -> int:
    """Run specified test steps"""

    if os.name == "nt":
        sys.stdout.reconfigure(encoding="utf-8")  # type: ignore

    if not version:
        proj = release.get_project(rt)
        version = str(proj.version)

    presets = Presets().visit_file(Path("CMakePresets.json")) or {}
    preset = presets.get(preset_name)

    if preset is None:
        print(f"error: preset `{preset_name}` not found", file=sys.stderr)
        return 1

    binary_dir = preset.expand()
    build_type = preset.build_type

    if not binary_dir:
        print(
            f"error: preset `{preset_name}` has no binaryDir attached to it",
            file=sys.stderr,
        )
        return 1

    if not build_type:
        print(
            f"error: preset `{preset_name}` has no CMAKE_BUILD_TYPE attached to it",
            file=sys.stderr,
        )
        return 1

    config = cast(dict, rt._cfg.get("test-runner", {}))
    target = cast(str | None, config.get("target"))

    if not isinstance(target, str):
        print(
            "error: cannot find test target; "
            "please add name of the executable as `target' property of "
            "`test-runner' config in flow's configuration file",
            file=sys.stderr,
        )
        return 1

    ext = ".exe" if sys.platform == "win32" else ""
    target_path = Path(binary_dir) / "bin" / (target + ext)
    if not target_path.is_file():
        print(
            f"error: cannot find {target + ext} in {target_path.parent.as_posix()}",
            file=sys.stderr,
        )
        return 1

    install_components = cast(list[str], config.get("install", []))
    patches = cast(dict[str, str], config.get("patches"))
    env_prefix = cast(str | None, config.get("report_env"))

    testsuite_config = cast(dict, config.get("testsuite", {}))
    test_root = cast(str | None, testsuite_config.get("root"))
    test_root_path = Path(test_root).resolve() if test_root else None
    if not test_root_path:
        print(
            "error: cannot find test root directory; "
            "please add name of the directory as `root' property of "
            "`test-runner/testsuite' config in flow's configuration file",
            file=sys.stderr,
        )
        return 1

    test_data_dir = cast(str | None, testsuite_config.get("data"))
    data_dir = (
        (test_root_path / test_data_dir).resolve()
        if isinstance(test_data_dir, str)
        else None
    )

    test_default_set = cast(str | None, testsuite_config.get("default-set"))
    if isinstance(test_default_set, str):
        if (
            not (test_root_path / tests).is_dir()
            and (test_root_path / test_default_set / tests).is_dir()
        ):
            tests = f"{test_default_set}/{tests}"

    test_set_dir = test_root_path / tests

    test_files = _enum_tests(test_set_dir, data_dir)
    tests_to_run = [int(x) for s in (run or []) for x in s.split(",")]
    if not tests_to_run:
        tests_to_run = list(range(1, len(test_files) + 1))

    independent_tests, linear_tests = _load_tests(test_files, tests_to_run)

    if nullify:
        for sequence in (independent_tests, linear_tests):
            for test in sequence:
                test[0].nullify(lang=None)
        return 0

    if not independent_tests and not linear_tests:
        print("No tests to run.", file=sys.stderr)
        return 0

    RUN_LINEAR = os.environ.get("RUN_LINEAR", 0) != 0
    if RUN_LINEAR:
        linear_tests[0:0] = independent_tests
        independent_tests = []

    try:
        thread_count = int(os.environ.get("POOL_SIZE", "not-a-number"))
    except (ValueError, TypeError):
        thread_count = max(1, (os.cpu_count() or 0)) * 2

    env = _make_env(
        target_path,
        data_dir or Path(binary_dir),
        version,
        len(independent_tests) + len(linear_tests),
        patches,
        env_prefix,
    )

    info_block: list[tuple[str, str]] = []

    info_block.append(("target", f"{env.target} {env.version}"))
    if env.data_dir_alt is None:
        info_block.append(("data", env.data_dir))
    else:
        info_block.append(("data", f"{env.data_dir} {env.data_dir_alt}"))
    info_block.append(("tests", tests))
    if env.tempdir_alt is None:
        info_block.append(("$TEMP", env.tempdir))
    else:
        info_block.append(("$TEMP", f"{env.tempdir} {env.tempdir_alt}"))
    if independent_tests:
        info_block.append(("threads", str(thread_count)))
    if sys.prefix != sys.base_prefix:
        info_block.append(
            (
                "python",
                f"{platform.python_version()} in {sys.prefix} ({sys.base_prefix})",
            )
        )
    else:
        info_block.append(("python", f"{platform.python_version()} in {sys.prefix}"))

    info_block.append(("proj-flow", __version__))
    info_block.append(("PyYAML", f"{yaml_version} (loader: {yaml_loader})"))

    _print_all(info_block)

    os.makedirs(env.tempdir, exist_ok=True)

    install_dir = Path("build").resolve() / ".test-runner"
    if not _install(
        install_dir,
        binary_dir,
        build_type,
        install_components,
        env,
    ):
        return 1

    return run_and_report_tests(
        independent_tests=independent_tests,
        linear_tests=linear_tests,
        install_dir=install_dir,
        env=env,
        thread_count=thread_count,
    )


def _enum_tests(test_root: Path, data_dir: Path | None):
    test_files: list[Path] = []
    for root, _, files in test_root.walk():
        for filename in files:
            if not Path(filename).suffix in [".json", ".yaml", ".yml"]:
                continue
            test_path = root / filename
            if data_dir and test_path.is_relative_to(data_dir):
                continue
            test_files.append(test_path)

    return test_files


def _load_tests(testsuite: list[Path], run: list[int]):
    counter: int = 0
    independent_tests: list[tuple[Test, int]] = []
    linear_tests: list[tuple[Test, int]] = []

    for filename in sorted(testsuite):
        counter += 1
        if counter not in run:
            continue

        try:
            test = Test.load(filename, counter)
            if not test.ok:
                continue
        except Exception:
            print(">>>>>>", filename.as_posix(), file=sys.stderr)
            raise

        if test.linear:
            linear_tests.append((test, counter))
        else:
            independent_tests.append((test, counter))

    return independent_tests, linear_tests


def _cmake_executable(binary_dir: str | PathLike[str]):
    cache = Path(binary_dir) / "CMakeCache.txt"
    if not cache.is_file():
        return "cmake"

    with cache.open(encoding="UTF-8") as cache_file:
        for line in cache_file:
            line = line.strip()
            if not line.startswith("CMAKE_COMMAND:INTERNAL="):
                continue
            return line.split("=", 1)[1]

    return "cmake"


def _run_with(test: Test, *args: str):
    cwd = None if test.linear else test.cwd
    subprocess.run(args, shell=False, cwd=cwd)


def _target(target: str):
    def impl(test: Test, aditional: list[str]):
        _run_with(test, target, *aditional)

    return impl


def _make_env(
    target: Path,
    data_dir: Path,
    version: str,
    counter_total: int,
    patches: dict[str, str],
    env_prefix: str | None,
):
    target_name = target.stem if os.name == "nt" else target.name
    tempdir = (Path(tempfile.gettempdir()) / "test-runner").resolve()
    tempdir_alt = None
    data_dir_alt = None

    if os.sep != "/":
        tempdir_alt = str(tempdir)
        data_dir_alt = str(data_dir)

    length = counter_total
    digits = 1
    while length > 9:
        digits += 1
        length = length // 10

    return Env(
        target=str(target),
        target_name=target_name,
        build_dir=str(target.parent.parent),
        data_dir=data_dir.as_posix(),
        inst_dir=str(data_dir.parent / "copy" / "bin"),
        tempdir=tempdir.as_posix(),
        version=version,
        counter_digits=digits,
        counter_total=counter_total,
        handlers=HANDLERS,
        data_dir_alt=data_dir_alt,
        tempdir_alt=tempdir_alt,
        builtin_patches=patches,
        reportable_env_prefix=env_prefix,
    )


def _install(
    dst: str | PathLike[str],
    binary_dir: str | PathLike[str],
    build_type: str,
    components: list[str],
    env: Env,
):
    dst = Path(dst)
    shutil.rmtree(dst, ignore_errors=True)
    dst.mkdir(parents=True, exist_ok=True)

    args = [
        _cmake_executable(binary_dir),
        "--install",
        str(binary_dir),
        "--config",
        build_type,
        "--prefix",
        str(dst),
    ]

    if not components:
        proc = subprocess.run(args, capture_output=True)
        return proc.returncode == 0

    args.append("--component")
    args.append("")

    for component in components:
        args[-1] = component
        proc = subprocess.run(args, capture_output=True)
        if proc.returncode != 0:
            return False

    env.target = str(dst / "bin" / os.path.basename(env.target))
    target_name = os.path.split(env.target)[1]
    if os.name == "nt":
        target_name = os.path.splitext(target_name)[0]
    env.handlers[target_name] = (0, _target(env.target))

    return True


def _print_all(info_block: list[tuple[str, str]]):
    length = 0
    for label, _ in info_block:
        label_len = len(label)
        if length < label_len:
            length = label_len
    length += 1

    for label, info in info_block:
        lab = f"{label}:"
        print(f"{lab: <{length}} {info}")



# Copyright (c) 2026 Marcin Zdun
# This code is licensed under MIT license (see LICENSE for details)

import typing
from pathlib import Path

import chevron

from proj_flow.api import arg, env
from proj_flow.ext.webidl.base.config import (
    CMakeDirs,
    TemplateConfig,
    load_package_template,
)
from proj_flow.ext.webidl.cli.updater import update_file_if_needed


@arg.command("webidl", "cmake")
def cmake(
    config_path: typing.Annotated[
        str, arg.Argument(help="Input configuration", names=["--cfg"], meta="json-file")
    ],
    binary_dir: typing.Annotated[
        str,
        arg.Argument(
            help="Project binary dir", names=["--binary-dir"], meta="project-binary-dir"
        ),
    ],
    target: typing.Annotated[
        str | None,
        arg.Argument(
            help="Name of the target to add output files to", meta="name", opt=True
        ),
    ],
    rt: env.Runtime,
):
    """Write the CMake configuration based on given config"""

    global_ctx = {
        key: f"${{{key}}}"
        for key in [
            "CMAKE_CURRENT_SOURCE_DIR",
            "CMAKE_CURRENT_BINARY_DIR",
            "PROJECT_SOURCE_DIR",
            "PROJECT_BINARY_DIR",
        ]
    }
    config_filename = Path(config_path)
    cmake_dirs = CMakeDirs.from_config_path(
        config_filename, project_binary_dir=Path(binary_dir)
    )

    abs_path = cmake_dirs.make_abs_path()

    depfile_ref = cmake_dirs.prefix(
        str(cmake_dirs.dependency_from_config(config_filename))
    )
    cmake_filename = cmake_dirs.cmake_from_config(config_filename)
    config_noext = cmake_dirs.filename_from_config(config_filename, "")

    config = TemplateConfig.load_config(
        config_filename.absolute(), global_ctx, cmake_dirs.project_source_dir, abs_path
    )

    context = {
        "config": {
            "binary": cmake_dirs.prefix(str(config_noext)).as_posix(),
            "source": cmake_dirs.prefix(str(config_filename.absolute())).as_posix(),
            "basename": config_filename.name,
        },
        "output": [
            out.output.as_posix() for rule in config.rules for out in rule.outputs
        ],
        "target": target,
        "depfile": depfile_ref.as_posix(),
    }

    text = load_package_template("cmake")
    text = chevron.render(text, data=context, partials_path=None)
    update_file_if_needed(
        cmake_filename, text, f"Generating {cmake_filename.as_posix()}"
    )



# Copyright (c) 2025 Marcin Zdun
# This code is licensed under MIT license (see LICENSE for details)

"""
The **proj_flow.minimal.base** provides basic initialization setup for all
new projects.
"""

import sys
from typing import List

from proj_flow import __version__, api
from proj_flow.flow import layer


class GitInit(api.init.InitStep):
    layers: List[layer.LayerInfo] = []

    def priority(self):
        return sys.maxsize

    def platform_dependencies(self):
        return ["git"]

    def postprocess(self, rt: api.env.Runtime, context: dict):
        def git(*args):
            rt.cmd("git", *args)

        git("init")
        git("add", ".")

        executables: List[str] = []

        for fs_layer in GitInit.layers:
            for info in fs_layer.get_git_checks():
                if info.is_executable:
                    executables.append(info.dst)

        if len(executables):
            git("update-index", "--chmod=+x", *executables)

        git("commit", "-m", "Initial commit")


api.init.register_init_step(GitInit())
api.ctx.register_init_setting(
    api.ctx.Setting("__flow_version__", value=__version__),
    api.ctx.Setting("${", value="${"),
    project=None,
    is_hidden=True,
)



# Copyright (c) 2026 Marcin Zdun
# This code is licensed under MIT license (see LICENSE for details)

import typing
from pathlib import Path

from proj_flow.api import arg, env
from proj_flow.ext.webidl.cli.updater import update_file_if_needed


@arg.command("webidl", "init")
def init(rt: env.Runtime):
    """Copy the main WebIDL CMake config to .flow/cmake"""

    out_dir_name = Path(".flow") / "cmake"
    in_dir_name = Path(__file__).parent.parent / "data" / "init"

    for root, _dirs, files in in_dir_name.walk():
        for filename in files:
            in_path = root / filename
            out_path = out_dir_name / in_path.relative_to(in_dir_name)
            update_file_if_needed(
                out_path,
                in_path.read_bytes().decode("utf-8"),
                f"adding {out_path.as_posix()}",
            )



# Copyright (c) 2025 Marcin Zdun
# This code is licensed under MIT license (see LICENSE for details)

"""
The **proj_flow.base.name_list** provides name list helper for arguments with
choices
"""

from typing import List


def name_list(names: List[str]) -> str:
    if len(names) == 0:
        return ""

    prefix = ", ".join(names[:-1])
    if prefix:
        prefix += " and "
    return f"{prefix}{names[-1]}"
